"""Holds the NexusStations class which is used to store and manipulate the stations in a NexusNetwork.

It is stored as an instance in the NexusNetwork class as "stations".
"""

from __future__ import annotations
from dataclasses import dataclass, field
from uuid import UUID
from typing import Literal, Sequence, Optional, TYPE_CHECKING

import pandas as pd

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.FileOperations.File import File
from ResSimpy.Nexus.nexus_add_new_object_to_file import AddObjectOperations
from ResSimpy.Nexus.nexus_collect_tables import collect_all_tables_to_objects
from ResSimpy.Nexus.DataModels.Network.NexusStation import NexusStation
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.nexus_modify_object_in_file import ModifyObjectOperations
from ResSimpy.Nexus.nexus_remove_object_from_file import RemoveObjectOperations
from ResSimpy.DataModelBaseClasses.OperationsMixin import NetworkOperationsMixIn
from ResSimpy.Utils.obj_to_dataframe import obj_to_dataframe

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusNetwork import NexusNetwork


@dataclass(kw_only=True)
class NexusStations(NetworkOperationsMixIn):
    """Class to store and manipulate the stations in a NexusStation.

    It is stored as an instance in the NexNexusStation usNetwork class as "stations".
    A list of stations in the network are stored in memory these can be accessed through
    the get_all method.
    """

    _stations: list[NexusStation] = field(default_factory=list)

    def __init__(self, parent_network: NexusNetwork) -> None:
        """Initialises the NexusStations class.

        Args:
            parent_network (NexusNetwork): The network that the stations are a part of.
        """
        self.__parent_network: NexusNetwork = parent_network
        self._stations: list[NexusStation] = []
        self.__add_object_operations = AddObjectOperations(
            NexusStation,
            self.table_header,
            self.table_footer,
            self.__parent_network.model,
        )
        self.__remove_object_operations = RemoveObjectOperations(
            self.__parent_network, self.table_header, self.table_footer
        )
        self.__modify_object_operations = ModifyObjectOperations(self)

    @property
    def table_header(self) -> str:
        """Start of the Station definition table."""
        return "STATION"

    @property
    def table_footer(self) -> str:
        """End of the Station definition table."""
        return "ENDSTATION"

    def get_all(self) -> Sequence[NexusStation]:
        """Returns a list of stations loaded from the simulator."""
        self.__parent_network.get_load_status()
        return self._stations

    def get_by_name(self, station_name: str) -> Optional[NexusStation]:
        """Returns a single station with the provided name loaded from the simulator.

        Args:
        ----
            station_name (str): name of the requested station

        Returns:
        -------
            NexusStation: which has the same name as the requested station_name

        """
        stations_to_return = filter(
            lambda x: (
                False if x.name is None else x.name.upper() == station_name.upper()
            ),
            self._stations,
        )
        return next(stations_to_return, None)

    def get_df(self) -> pd.DataFrame:
        """Creates a dataframe representing all processed station data in a surface file.

        Returns:
            DataFrame: of the properties of the stations through time with each row representing a station.
        """
        self.__parent_network.get_load_status()
        df_store = obj_to_dataframe(self._stations)
        return df_store

    def get_overview(self) -> str:
        """Returns overview of the stations."""
        raise NotImplementedError("To be implemented")

    def load(
        self, surface_file: File, start_date: str, default_units: UnitSystem
    ) -> None:
        """Calls load stations and appends the list of discovered stations into the NexusStations object.

        Args:
            surface_file (File): NexusFile representation of the surface file.
            start_date (str): Starting date of the run
            default_units (UnitSystem): Units used in case not specified by surface file.

        Raises:
            TypeError: if the unit system found in the property check is not a valid enum UnitSystem.

        """
        new_stations, _ = collect_all_tables_to_objects(
            surface_file,
            {"STATIONS": NexusStation},
            start_date=start_date,
            default_units=default_units,
            date_format=self.__parent_network.model.date_format,
        )
        stations_list = new_stations.get("STATIONS")
        self._add_to_memory(stations_list)

    def _add_to_memory(self, additional_list: Optional[list[NexusStation]]) -> None:
        """Extends the stations object by a list of stations provided to it.

        Args:
        ----
            additional_list (Sequence[NexusStation]): list of nexus stations to add to the stations list.

        Returns:
        -------
            None
        """
        if additional_list is None:
            return
        self._stations.extend(additional_list)

    def remove(
        self, station_to_remove: DataObjectMixinDictType | UUID
    ) -> None:
        """Remove a station from the network based on the properties matching a dictionary or id.

        Args:
            station_to_remove (UUID | DataObjectMixinDictType): UUID of the station to remove
            or a dictionary with sufficient matching parameters to uniquely identify a station

        """
        self.__remove_object_operations.remove_object_from_network_main(
            station_to_remove, self._network_element_name, self._stations
        )

    def add(self, station_to_add: DataObjectMixinDictType) -> None:
        """Adds a station to a network, taking a dictionary with properties for the new station.

        Args:
            station_to_add (DataObjectMixinDictType): dictionary taking all the properties for new station.
            Requires date and a station name.
        """
        new_object = self.__add_object_operations.add_network_obj(
            station_to_add, NexusStation, self.__parent_network
        )
        self._add_to_memory([new_object])

    def modify(
        self,
        station_to_modify: DataObjectMixinDictType,
        new_properties: DataObjectMixinDictType,
    ) -> None:
        """Modifies an existing station based on a matching dictionary of properties.

        Partial matches allowed if precisely 1 matching station is found. Updates the properties with properties in the
        new_properties dictionary.

        Args:
            station_to_modify (DataObjectMixinDictType): dictionary containing attributes to match in the
            existing station set.
            new_properties (DataObjectMixinDictType): properties to switch to in the new station
        """
        self.__parent_network.get_load_status()

        self.__modify_object_operations.modify_network_object(
            station_to_modify, new_properties, self.__parent_network
        )

    @property
    def _network_element_name(self) -> Literal["stations"]:
        return "stations"



# TODO: Not a complete list. Check manual for missing keywords.
# TODO: Delete the keywords that are not tokens (i.e. that are only values in a table)

OPT_KEYWORDS_VALUE_FLOAT = ['PSTD', 'TSTD', 'RES_TEMP', 'MWOIL', 'MWGAS', 'MWOIL_L', 'MWOIL_H', 'DEF_DATUM',
                            'FUG_XY_TOL']
OPT_SINGLE_KEYWORDS = ['IRF_REPORT', 'LGRSTATIC', 'ZOLTAN']
OPT_ARRAY_KEYWORDS = ['BOUNDARY_FLUXIN_SECTORS']
OPT_GLOBAL_METHOD_OVERRIDES_KEYWORDS = ['NONEQ', 'STONE1', 'STONE2', 'KROINT',
                                        'NOCHK_HYS', 'STONE1_WAT', 'STONE2_WAT', 'KRWINT',
                                        'VIP_INJ_PERF_KR_SCALING', 'NOCHK_SAL_TEMP']

OTHER_OPTIONS_KEYWORDS = ['ALL', 'AQUIFER', 'AUTO', 'CENTER', 'DATUM', 'DEF_DATUM', 'DESC', 'DP',
                          'ENDGLOBAL_METHOD_OVERRIDES', 'ENDGRIDTOPROC', 'ENDREGBLK', 'ENDREGDATA', 'ENDSUBREGIONS',
                          'EQ_LIST', 'FLEXDP', 'GLOBAL_METHOD_OVERRIDES', 'GRID', 'GRIDBLOCKS', 'GRIDTOPROC', 'INTSAT',
                          'NAME', 'NOCHK_SAL_TEMP', 'NUMBER', 'OFF', 'OVERREAD', 'PATTERN', 'PC', 'PD',
                          'POROSITY_INDEPENDENCE', 'PROCESS', 'REGBLK', 'REGDATA', 'RESERVOIR', 'SOMOPT1',
                          'STONE1', 'SUBREGION', 'SUBREGIONS', 'SW', 'VIP_INJ_PERF_KR_SCALING']

OPTIONS_KEYWORDS = OPT_KEYWORDS_VALUE_FLOAT + OPT_SINGLE_KEYWORDS + OPT_GLOBAL_METHOD_OVERRIDES_KEYWORDS \
    + OTHER_OPTIONS_KEYWORDS



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Mapping

from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty


@dataclass
class DynamicPropertyContainer(ABC):
    """Class for handling a collection of dynamic properties, such as PVT, relperm, etc."""

    @property
    @abstractmethod
    def inputs(self) -> Mapping[int, DynamicProperty]:
        """A Collection of dynamic properties, as a dictionary, keyed by method number."""
        raise NotImplementedError("Implement this in the derived class.")



"""Holds a class for represeting a single NodeList or group for the Nexus model."""

from dataclasses import dataclass

from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Units.AttributeMappings.BaseUnitMapping import BaseUnitMapping
from ResSimpy.DataModelBaseClasses.NetworkList import NetworkList


@dataclass(kw_only=True, init=False)
class NexusNodeList(NetworkList):
    """Class for representing a single NodeList or group for the Nexus model."""

    def __init__(
        self,
        name: str,
        elements_in_the_list: list[str],
        date: str,
        date_format: DateFormat,
    ) -> None:
        """Initialises the NexusNodeList class.

        Args:
            name (str): Name of the NodeList.
            elements_in_the_list (list[str]): List of node names in the NodeList.
            date (str): Date when the NodeList is defined. Persists until the next date is defined.
            date_format (Optional[DateFormat]): The date format of the object.
        """
        super().__init__(
            name=name,
            elements_in_the_list=elements_in_the_list,
            date=date,
            date_format=date_format,
        )

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Returns mapping of keywords."""
        return {}

    @property
    def units(self) -> BaseUnitMapping:
        """Returns the attribute to unit map for the data object."""
        return BaseUnitMapping(None)

    @property
    def nodes(self) -> list[str]:
        """Returns the list of nodes in the NodeList."""
        return self.elements_in_the_list



"""Class for representing a well in Nexus. Consists of a list of completions and a well name."""
from __future__ import annotations
from dataclasses import dataclass, field
from typing import Optional, Sequence, TYPE_CHECKING
from uuid import UUID

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.Enums.WellTypeEnum import WellType
from ResSimpy.Nexus.DataModels.NexusCompletion import NexusCompletion
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.DataModels.NexusWellMod import NexusWellMod
from ResSimpy.Time.ISODateTime import ISODateTime
from ResSimpy.Utils.generic_repr import generic_repr, generic_str
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.DataModelBaseClasses.Well import Well
from ResSimpy.Utils.obj_to_table_string import get_column_headers_required

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusWells import NexusWells


@dataclass(kw_only=True)
class NexusWell(Well):
    """Class for representing a well in Nexus. Consists of a list of completions and a well name."""
    _wellmods: list[NexusWellMod]
    _parent_wells_instance: NexusWells = field(compare=False, repr=False)

    def __init__(self, well_name: str, completions: Sequence[NexusCompletion], unit_system: UnitSystem,
                 parent_wells_instance: NexusWells, wellmods: Sequence[NexusWellMod] | None = None,
                 well_type: Optional[WellType] = None) -> None:
        """Initialises the NexusWell class.

        Args:
            well_name (str): Name of the well
            completions (list[NexusCompletion]): List of completions for the well.
            unit_system (UnitSystem): Unit system for the well.
            parent_wells_instance (NexusWells): Instance of the parent wells class that the well belongs to.
            wellmods (Sequence[NexusWellMod] | None): List of wellmods for the well. Defaults to None.
            well_type (Optional[WellType]): Type of the well as an enum. Defaults to None.
        """
        self._parent_wells_instance = parent_wells_instance

        if wellmods is None:
            wellmods = []
        elif not isinstance(wellmods, list):
            wellmods = list(wellmods)
        self._wellmods: list[NexusWellMod] = wellmods
        well_type = WellType.PRODUCER if well_type is None else well_type
        super().__init__(well_name=well_name, completions=completions, unit_system=unit_system, well_type=well_type)

    def __repr__(self) -> str:
        return generic_repr(self, exclude_attributes=['_parent_wells_instance'])

    def __str__(self) -> str:
        return generic_str(self)

    @property
    def well_type(self) -> WellType | None:
        """The type of the well."""

        # Ensure that the wells have been populated with the information from the network as well
        self._parent_wells_instance.model.network.get_load_status()
        return self._well_type

    @well_type.setter
    def well_type(self, val: WellType) -> None:
        """Sets the well type."""
        if not isinstance(val, WellType):
            raise ValueError(f"Invalid well type: {val}")

        self._well_type = val

    @property
    def wellmods(self) -> list[NexusWellMod]:
        """Property with a list of all the wellmods for the well."""
        return self._wellmods

    def find_completions(self, completion_properties: DataObjectMixinDictType | NexusCompletion) -> \
            list[NexusCompletion]:
        """Returns a list of all completions that match the completion properties provided.

        Args:
        ----
            completion_properties (dict | NexusCompletion): keys as the attributes and values as the value to match \
            from the well.

        Returns:
        -------
            list[NexusCompletion] that match the completion properties provided
        """
        matching_completions: list[NexusCompletion] = []
        if isinstance(completion_properties, NexusCompletion):
            perf_props = completion_properties.to_dict(add_units=False)
        else:
            perf_props = completion_properties
        perf_props_without_nones = {k: v for k, v in perf_props.items() if v is not None}
        for i, perf in enumerate(self._completions):
            for prop, value in perf_props_without_nones.items():
                if getattr(perf, prop) == value:
                    # go to the next perf if a value from the dictionary doesn't match
                    continue
                else:
                    break
            else:
                if isinstance(perf, NexusCompletion):
                    # if all the conditions match then append the perf to completions
                    matching_completions.append(perf)
        return matching_completions

    def find_completion(self,
                        completion_properties: DataObjectMixinDictType | NexusCompletion) \
            -> NexusCompletion:
        """Returns precisely one completion which matches all the properties provided.

            If several completions match it raise a ValueError.

        Args:
        ----
            completion_properties (dict | NexusCompletion): The completion properties.

        Returns:
        -------
            NexusCompletion that matches all completion properties provided.
        """
        matching_completions = self.find_completions(completion_properties)
        if len(matching_completions) != 1:
            raise ValueError(f'Could not find single unique completion that matches property provided. Instead found: '
                             f'{len(matching_completions)} completions')
        return matching_completions[0]

    def _add_completion_to_memory(self, date: str, completion_properties: DataObjectMixinDictType,
                                  date_format: DateFormat, completion_index: Optional[int] = None) -> NexusCompletion:
        """Adds a perforation with the properties specified in completion_properties_list.

            If index is none then adds it to the end of the perforation list.

        Args:
            date (str): date at which the perforation should be added
            completion_properties (dict[str, str | float | int]): The completion properties.
            date_format (DateFormat): The date format.
            completion_index (Optional[int]): The index of the completion being searched for.
        """
        completion_properties['date'] = date
        completion_properties['unit_system'] = self.unit_system
        completion_properties['start_date'] = self._parent_wells_instance.start_date
        new_completion = NexusCompletion.from_dict(completion_properties, date_format)
        if completion_index is None:
            completion_index = len(self._completions)
        self._completions.insert(completion_index, new_completion)
        return new_completion

    def _remove_completion_from_memory(self, completion_to_remove: NexusCompletion | UUID) -> None:
        if isinstance(completion_to_remove, NexusCompletion):
            completion_to_remove = self.find_completion(completion_to_remove)
            completion_to_remove = completion_to_remove.id
        completion_index_to_remove = [x.id for x in self._completions].index(completion_to_remove)
        self._completions.pop(completion_index_to_remove)

    def _modify_completion_in_memory(self, new_completion_properties: DataObjectMixinDictType,
                                     completion_to_modify: NexusCompletion | UUID,
                                     ) -> None:
        if isinstance(completion_to_modify, NexusCompletion):
            completion = self.find_completion(completion_to_modify)
        else:
            completion = self.get_completion_by_id(completion_to_modify)
        completion.update(new_completion_properties)

    def get_completion_by_id(self, id: UUID) -> NexusCompletion:
        """Returns the completion that matches the id provided."""
        for completion in self._completions:
            if completion.id == id and isinstance(completion, NexusCompletion):
                return completion
        raise ValueError('No completion found for id: {id}')

    def _modify_completions_in_memory(self, new_completion_properties: DataObjectMixinDictType,
                                      completions_to_modify: list[NexusCompletion | UUID]) -> None:
        for completion in completions_to_modify:
            if isinstance(completion, UUID):
                modify_this_completion = self.get_completion_by_id(completion)
                modify_this_completion.update(new_completion_properties)
            else:
                completion.update(new_completion_properties)

    def _remove_completions_from_memory(self, completions_to_remove: Sequence[NexusCompletion | UUID]) -> None:
        # TODO improve comparison of dates with datetime libs
        """Removes completions from a list of completions or completion IDs.

        Args:
            completions_to_remove (list[NexusCompletion | UUID]): Completions to be removed.
        """
        for completion in completions_to_remove:
            self._remove_completion_from_memory(completion)

    def to_string_for_date(self, date: ISODateTime) -> str:
        """Outputs the string that represents the wellspec for a specific date."""
        output_str = ''
        completions_for_date = [x for x in self._completions if x.iso_date == date]
        wellmods_for_date = [x for x in self._wellmods if x.iso_date == date]

        if not completions_for_date and not wellmods_for_date:
            return output_str

        if completions_for_date:
            output_str += f"WELLSPEC {self.well_name}\n"
            # work out the required headers:
            headers = get_column_headers_required(completions_for_date)
            output_str += ' '.join(headers) + '\n'
            for completion in completions_for_date:
                output_str += completion.to_table_line(headers)

        if wellmods_for_date:
            for wellmod in wellmods_for_date:
                output_str += wellmod.to_string()

        output_str += '\n'
        return output_str



"""Class to hold Nexus PVT properties."""
from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Union
import numpy as np
import pandas as pd
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.NexusKeywords.pvt_keywords import PVT_BLACKOIL_PRIMARY_KEYWORDS, PVT_TYPE_KEYWORDS, PVT_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.pvt_keywords import PVT_EOS_METHODS, PVT_EOSOPTIONS_PRIMARY_WORDS
from ResSimpy.Nexus.NexusKeywords.pvt_keywords import PVT_EOSOPTIONS_PRIMARY_KEYS_INT, PVT_TABLE_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.pvt_keywords import PVT_TABLES_WITH_ENDWORDS, PVT_TABLES_WITHOUT_ENDWORDS
from ResSimpy.Nexus.NexusKeywords.pvt_keywords import PVT_EOSOPTIONS_PRIMARY_KEYS_FLOAT, PVT_EOSOPTIONS_TRANS_KEYS
from ResSimpy.Nexus.NexusKeywords.pvt_keywords import PVT_EOSOPTIONS_TRANS_TEST_KEYS, PVT_EOSOPTIONS_PHASEID_KEYS
from ResSimpy.Nexus.NexusKeywords.pvt_keywords import PVT_EOSOPTIONS_TERTIARY_KEYS, PVT_ALL_TABLE_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.pvt_keywords import PVT_UNSAT_TABLE_INDICES
from ResSimpy.Enums.FluidTypeEnums import PvtType
from ResSimpy.Enums.UnitsEnum import UnitSystem, SUnits, TemperatureUnits
from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.Units.AttributeMappings.DynamicPropertyUnitMapping import PVTUnits

from ResSimpy.Utils.factory_methods import get_empty_dict_union, get_empty_list_str
from ResSimpy.Utils.factory_methods import get_empty_eosopt_dict_union
import ResSimpy.Nexus.nexus_file_operations as nfo
import ResSimpy.FileOperations.file_operations as fo


@dataclass(kw_only=True, repr=False)
class NexusPVTMethod(DynamicProperty):
    """Class to hold Nexus PVT properties.

    Attributes:
        file (NexusFile): Nexus PVT file object
        input_number (int): PVT method number in Nexus fcs file
        pvt_type (Optional[str]): Type of PVT method, e.g., BLACKOIL, GASWATER or EOS. Defaults to None
        eos_nhc (Optional[int]): Number of hydrocarbon components. Defaults to None
        eos_temp (Optional[float]): Default temperature for EOS method. Defaults to None
        eos_components (Optional[list[str]]): Specifies component names
        eos_options (dict[str, Union[str, int, float, pd.DataFrame, list[str], dict[str, float],
            tuple[str, dict[str, float]], dict[str, pd.DataFrame]]]): Dictionary containing various EOS options
            as specified in the PVT file. Defaults to empty dictionary.
        properties (dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                    dict[str, Union[float, pd.DataFrame]]]]):
            Dictionary holding all properties for a specific PVT method. Defaults to empty dictionary.
    """

    # General parameters
    file: NexusFile = field(compare=False)
    pvt_type: Optional[PvtType] = None
    eos_nhc: Optional[int] = None  # Number of hydrocarbon components
    eos_temp: Optional[float] = None  # Default temperature for EOS method
    eos_components: Optional[list[str]] = field(default_factory=get_empty_list_str)
    eos_options: dict[str, Union[
        str, int, float, pd.DataFrame, list[str], dict[str, float], tuple[str, dict[str, float]], dict[
            str, pd.DataFrame]]] \
        = field(default_factory=get_empty_eosopt_dict_union)
    properties: dict[str, Union[str, int, float, Enum, list[str], np.ndarray,
                                pd.DataFrame, dict[str, Union[float, pd.DataFrame]]]] \
        = field(default_factory=get_empty_dict_union)
    unit_system: UnitSystem

    def __init__(self, file: NexusFile, input_number: int, model_unit_system: UnitSystem,
                 pvt_type: Optional[PvtType] = None,
                 eos_nhc: Optional[int] = None, eos_temp: Optional[float] = None,
                 eos_components: Optional[list[str]] = None,
                 eos_options: Optional[dict[str, Union[str, int, float, pd.DataFrame, list[str], dict[str, float],
                                                       tuple[str, dict[str, float]], dict[str, pd.DataFrame]]]] = None,
                 properties: Optional[dict[str, Union[str, int, float, Enum, list[str], np.ndarray, pd.DataFrame,
                                                      dict[str, Union[float, pd.DataFrame]]]]] = None) -> None:
        """Initialises the NexusGasliftMethod class.

        Args:
            file (NexusFile): NexusFile object associated with the PVT method.
            input_number (int): method number for the PVT method.
            model_unit_system (UnitSystem): unit system from the model.
            pvt_type (Optional[PvtType]): Type of PVT method, e.g., BLACKOIL, GASWATER or EOS. Defaults to None.
            eos_nhc (Optional[int]): Number of hydrocarbon components for EOS method. Defaults to None.
            eos_temp (Optional[float]): Default temperature for EOS method. Defaults to None.
            eos_components (Optional[list[str]]): Specifies component names for EOS method. Defaults to None.
            eos_options: Dictionary containing various EOS options as specified in the PVT file. Defaults to None.
            properties: dictionary of properties for the PVT method.
        """
        if pvt_type is not None:
            self.pvt_type = pvt_type
        if eos_nhc is not None:
            self.eos_nhc = eos_nhc
        if eos_temp is not None:
            self.eos_temp = eos_temp
        if eos_components is not None:
            self.eos_components = eos_components
        else:
            self.eos_components = []
        if eos_options is not None:
            self.eos_options = eos_options
        else:
            self.eos_options = {}
        if properties is not None:
            self.properties = properties
        else:
            self.properties = {}
        self.unit_system = model_unit_system
        super().__init__(input_number=input_number, file=file)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords: dict[str, tuple[str, type]] = {
            'PRES': ('pressure', float),
            'TEMP': ('temperature', float),
            'DP': ('delta_pressure', float),
            'DENOIL': ('oil_density', float),
            'DENGAS': ('gas_density', float),
            'API': ('oil_api_gravity', float),
            'SPECG': ('gas_specific_gravity', float),
            'MWOR': ('molecular_weight_of_residual_oil', float),
            'VO': ('oil_viscosity', float),
            'VG': ('gas_viscosity', float),
            'BO': ('oil_formation_volume_factor', float),
            'BG': ('gas_formation_volume_factor', float),
            'RS': ('solution_gas_oil_ratio', float),
            'RV': ('solution_oil_gas_ratio', float),
            'PSAT': ('saturation_pressure', float),
            'GOR': ('gas_oil_ratio', float),
            'OGR': ('oil_gas_ratio', float),
            'CV': ('gas_heat_capacity_at_constant_volume', float),
            'CP': ('gas_heat_capacity_at_constant_pressure', float),
            'BOFAC': ('oil_fvf_over_oil_fvf_at_saturation_pressure', float),
            'VOFAC': ('oil_viscosity_over_oil_viscosity_at_saturation_pressure', float),
            'BGFAC': ('gas_fvf_over_gas_fvf_at_saturation_pressure', float),
            'VGFAC': ('gas_viscosity_over_gas_viscosity_at_saturation_pressure', float),
            'CVFAC': ('gas_heat_capacity_at_constant_volume_over_cv_at_saturation_pressure', float),
            'CPFAC': ('gas_heat_capacity_at_constant_pressure_over_cp_at_saturation_pressure', float),
            'PC': ('critical_pressure', float),
            'TC': ('critical_temperature', float),
            'VC': ('critical_volume', float),
        }
        return keywords

    @staticmethod
    def __hash_props(props: Optional[dict[str, Union[str, int, float, pd.DataFrame, list[str], dict[str, float],
                                                     tuple[str, dict[str, float]], dict[str, pd.DataFrame], np.ndarray,
                                                     dict[str, Union[float, pd.DataFrame]]]]]) -> int:
        """Converts properties and EOS options to a hashable format."""
        if props is None:
            return hash(None)

        hashable_dict = frozenset((k, DynamicProperty.convert_to_hashable(v)) for k, v in props.items())
        return hash(hashable_dict)

    def __hash__(self) -> int:
        """Hash the PVT method object."""
        eos_options_hash = self.__hash_props(self.eos_options)
        eos_components = None if self.eos_components is None else tuple(self.eos_components)
        eos_attributes = (self.eos_nhc, self.eos_temp, eos_components, eos_options_hash)
        properties_hash = self.__hash_props(self.properties)
        # add the units to the hash
        unit_system_hash = hash(self.unit_system)
        return hash((eos_attributes, properties_hash, unit_system_hash))

    @property
    def units(self) -> PVTUnits:
        """Returns the attribute to unit map for the PVT method."""
        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']
        return PVTUnits(unit_system=self.unit_system)

    def to_string(self) -> str:
        """Create string with PVT data in Nexus file format."""
        printable_str = ''
        pvt_dict = self.properties
        # Print description if present
        if 'DESC' in pvt_dict.keys() and isinstance(pvt_dict['DESC'], list):
            for desc_line in pvt_dict['DESC']:
                printable_str += 'DESC ' + desc_line + '\n'
        # Print PVT type and associated properties
        printable_str += '' if self.pvt_type is None else f'{self.pvt_type.value}'
        if self.eos_nhc is not None:
            printable_str += f' NHC {self.eos_nhc}'
        for pvt_key in PVT_BLACKOIL_PRIMARY_KEYWORDS:
            if pvt_key in pvt_dict.keys():
                printable_str += f' {pvt_key} {pvt_dict[pvt_key]}'
        printable_str += '\n'
        if 'DRYGAS_MFP' in pvt_dict.keys():
            printable_str += 'DRYGAS_MFP\n'
        # Print EOS-specific required parameters, if present
        if self.eos_components is not None:
            if len(self.eos_components) > 0:
                printable_str += f"COMPONENTS {' '.join(self.eos_components)}\n"
        if self.eos_temp is not None:
            printable_str += f'TEMP {self.eos_temp}\n'
        for key, value in pvt_dict.items():
            if isinstance(value, pd.DataFrame):
                printable_str += f'{key}\n'
                if key == 'IGS_CP':
                    printable_str += value.to_string(na_rep='', index=False, header=False) + '\n'
                else:
                    printable_str += value.to_string(na_rep='', index=False,
                                                     float_format=lambda x: f'{x:.9f}') + '\n'
                if key in PVT_TABLES_WITH_ENDWORDS:
                    printable_str += 'END' + key + '\n'
                printable_str += '\n'
            elif isinstance(value, dict):
                for subkey in value.keys():
                    printable_str += f"{key.replace('_', ' ')} {subkey}\n"
                    df = value[subkey]
                    if isinstance(df, pd.DataFrame):
                        printable_str += df.to_string(na_rep='', index=False,
                                                      float_format=lambda x: f'{x:.9f}') + '\n'
                    printable_str += '\n'
            elif isinstance(value, Enum):
                if isinstance(value, UnitSystem) or isinstance(value, TemperatureUnits):
                    printable_str += f'{value.value}\n'
                elif isinstance(value, SUnits):
                    printable_str += f'SUNITS {value.value}\n'
            elif key not in [*PVT_BLACKOIL_PRIMARY_KEYWORDS, 'NHC', 'DESC', 'DRYGAS_MFP']:
                if value == '':
                    printable_str += f'{key}\n'
                else:
                    printable_str += f'{key} {value}\n'
        if len(self.eos_options.keys()) > 0:
            pvt_dict = self.eos_options
            printable_str += 'EOSOPTIONS'
            if 'EOS_METHOD' in pvt_dict.keys():
                printable_str += f" {pvt_dict['EOS_METHOD']}"
            printable_str += '\n'
            for key, value in pvt_dict.items():
                if key == 'EOS_OPT_PRIMARY_LIST' and isinstance(value, list):
                    for token in value:
                        printable_str += f'    {token}\n'
                elif isinstance(value, tuple):
                    printable_str += f'    {key} {value[0]}'
                    val_dict: dict[str, float] = value[1]
                    for val_key, val_val in val_dict.items():
                        printable_str += f' {val_key} {val_val}'
                    printable_str += '\n'
                elif isinstance(value, dict):
                    printable_str += f'    {key}'
                    for subkey, subvalue in value.items():
                        printable_str += f' {subkey} {subvalue}'
                    printable_str += '\n'
                elif key not in ['EOS_METHOD']:
                    printable_str += f'    {key} {value}\n'
        return printable_str

    def __populate_eos_opts_to_tertiary_keys(self, primary_key: str, primary_key_default_val: str, single_line: str,
                                             line_list: list[str], list_of_secondary_keys: list[str]) -> None:
        """Utility function to populate complex EOS options structures, from primary to tertiary keyword level.

        Applies to TRANSITION, TRANS_TEST and PHASEID Nexus EOS options.

        Args:
            primary_key (str): primary keyword, e.g., TRANSITION or PHASEID
            primary_key_default_val (str): default secondary keyword, or primary key value, e.g., TEST
            single_line (str): single line as read from input PVT file
            line_list (list[str]): list of strings that comprise input PVT file
            list_of_secondary_keys (list[str]): list of secondary keywords associated with the given primary keyword
        """
        if nfo.check_token(primary_key, single_line):
            self.eos_options[primary_key] = primary_key_default_val  # Set default value
            if nfo.get_expected_token_value(primary_key, single_line, line_list) in list_of_secondary_keys:
                self.eos_options[primary_key] = nfo.get_expected_token_value(primary_key, single_line, line_list)
        if [i for i in single_line.split() if i in list_of_secondary_keys]:
            for secondary_key in list_of_secondary_keys:
                if nfo.check_token(secondary_key, single_line):
                    self.eos_options[primary_key] = secondary_key
                    for tertiary_key in PVT_EOSOPTIONS_TERTIARY_KEYS:
                        if nfo.check_token(tertiary_key, single_line):
                            if isinstance(self.eos_options[primary_key], str):  # Convert to tuple
                                self.eos_options[primary_key] = (secondary_key, {})

                            secondary_eos_option = self.eos_options[primary_key]
                            if not isinstance(secondary_eos_option, tuple) or \
                                    not isinstance(secondary_eos_option[1], dict):
                                raise ValueError(f"EOS secondary key invalid: {secondary_key}")
                            secondary_eos_option[1][tertiary_key] = float(
                                nfo.get_expected_token_value(tertiary_key, single_line, line_list))

    def __find_pvt_table_starting_index(self, table_key: str, single_line: str, line_list: list[str],
                                        table_indices: dict[str, list[int]],
                                        table_indices_dict: dict[str, dict[str, list[int]]],
                                        table_flag: dict[str, bool], l_index: int,
                                        unsat_obj: dict[str, list[str]] | None = None) -> Optional[int]:
        """Utility function to find the starting line index of a specified PVT table.

        Args:
            table_key (str): specified PVT table name or undersaturated index, such as, PSAT or RSSAT or PRES
            single_line (str): single line as read from input PVT file
            line_list (list[str]): list of strings that comprise input PVT file
            table_indices ([dict[str, list[int]]): dictionary to store the
                starting and ending line index of tables
            table_indices_dict (dict[str, dict[str, list[int]]]): dictionary to store the
                starting and ending line index of tables, for undersaturated tables
            table_flag (bool): flag to tell if a table is currently being read (True) or not (False)
            l_index (int): current line index
            unsat_obj (dict[str, list[str]]): track saturation pressures from which undersaturated branches emanate

        Raises:
            ValueError: If a property table key does not have a numerical value

        Returns:
            int: Updated line index
        """
        if unsat_obj is None:
            unsat_obj = {}
        if table_key not in PVT_UNSAT_TABLE_INDICES:  # All tables except undersaturated tables
            if nfo.check_token(table_key, single_line):
                table_indices[table_key] = [l_index + 1, len(line_list)]
                table_flag[table_key] = True
                return l_index + 1
        else:  # Handle undersaturated tables
            if table_key == 'PRES':
                table_name = 'UNSATGAS'
                full_table_name = table_name + '_PRES'
            else:
                table_name = 'UNSATOIL'
                full_table_name = table_name + '_' + table_key
            if nfo.check_token(table_name, single_line) and nfo.check_token(table_key, single_line):
                if fo.get_token_value(table_key, single_line, line_list) is None:
                    raise ValueError(f'Property {table_key} does not have a numerical value.')
                unsat_obj[table_key].append(nfo.get_expected_token_value(table_key, single_line, line_list))
                if full_table_name in table_indices_dict.keys():
                    table_indices_dict[full_table_name][unsat_obj[table_key][-1]] = [l_index + 1, len(line_list)]
                else:
                    table_indices_dict[full_table_name] = {unsat_obj[table_key][-1]: [l_index + 1, len(line_list)]}
                table_flag[table_name] = True
                return l_index + 1
        return None

    def __find_pvt_table_ending_index(self, table_key: str, single_line: str,
                                      table_indices: dict[str, list[int]],
                                      table_indices_dict: dict[str, dict[str, list[int]]],
                                      l_index: int, table_flag: dict[str, bool],
                                      table_has_endkeyword: bool, reading_flag: bool,
                                      unsat_obj: dict[str, list[str]] = {}) -> bool:
        """Utility function to find the ending line index of a specified PVT table.

        Args:
            table_key (str): specified PVT table name or undersaturated index, such as, PSAT or RSSAT or PRES
            single_line (str): single line as read from input PVT file
            table_indices ([dict[str, list[int]]): dictionary to store the
                starting and ending line index of tables
            table_indices_dict (dict[str, dict[str, list[int]]]): dictionary to store the
                starting and ending line index of tables, for undersaturated tables
            l_index (int): current line index
            table_flag (dict[str, bool]): flag to tell if a table is currently being read (True) or not (False)
            table_has_endkeyword (bool): True if table name, e.g., PROPS, has end keyword, i.e., ENDPROPS, else False
            reading_flag (bool): True if any table is being read, otherwise False
            unsat_obj (dict[str, list[str]]): track saturation pressures from which undersaturated branches emanate

        Returns:
            bool: True if still reading table, but if identified the ending line index, return False
        """
        end_flag_found = False
        if table_key not in PVT_UNSAT_TABLE_INDICES:  # All tables except undersaturated tables
            table_name = table_key
            full_table_name = table_key
        else:  # Handle undersaturated tables
            if table_key == 'PRES':
                table_name = 'UNSATGAS'
                full_table_name = table_name + '_PRES'
            else:
                table_name = 'UNSATOIL'
                full_table_name = table_name + '_' + table_key
        # if not table_has_endkeyword and [i for i in single_line.split() if i in PVT_KEYWORDS]:
        if not table_has_endkeyword:
            for keyword in PVT_KEYWORDS:
                if nfo.check_token(keyword, single_line):
                    end_flag_found = True
        if table_has_endkeyword and nfo.check_token('END' + table_name, single_line):
            end_flag_found = True
        if (full_table_name in table_indices.keys() or full_table_name in table_indices_dict.keys()) and \
                end_flag_found and table_flag[table_name]:
            if table_key not in PVT_UNSAT_TABLE_INDICES:  # All tables except undersaturated tables
                table_indices[table_name][1] = l_index
            else:  # Handle undersaturated tables
                table_indices_dict[full_table_name][unsat_obj[table_key][-1]][1] = l_index
            table_flag[table_name] = False
            reading_flag = False
        return reading_flag

    def read_properties(self) -> None:
        """Read Nexus PVT file contents and populate the NexusPVTMethod object."""
        file_as_list = self.file.get_flat_list_str_file

        # Check for common input data
        nfo.check_for_and_populate_common_input_data(file_as_list, self.properties)

        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']

        # Initialize flags and containers used to record properties as we iterate through pvt file contents
        # Dictionary to record start and ending indices for tables
        pvt_table_indices: dict[str, list[int]] = {}
        pvt_table_indices_dict: dict[str, dict[str, list[int]]] = {}
        # Flag to tell when to start reading a table
        start_reading_table: bool = False
        trans_flag = False
        # Dictionary of flags indicating which tables are being read
        table_being_read: dict[str, bool] = {}
        for table_name in PVT_ALL_TABLE_KEYWORDS:
            table_being_read[table_name] = False
        # Dictionary of lists to track saturation pressures from which undersaturated branches emanate
        pvt_unsat_keys: dict[str, list[str]] = {}
        for indx in PVT_UNSAT_TABLE_INDICES:
            pvt_unsat_keys[indx] = []

        line_indx = 0
        for line in file_as_list:

            # Determine PVT type, i.e., BLACKOIL, WATEROIL, EOS, etc.
            for pvt_type in PVT_TYPE_KEYWORDS:
                if nfo.check_token(pvt_type, line):
                    self.pvt_type = PvtType[pvt_type]

            # Extract blackoil fluid density parameters
            for fluid_param in PVT_BLACKOIL_PRIMARY_KEYWORDS:
                if nfo.check_token(fluid_param, line):
                    self.properties[fluid_param] = float(nfo.get_expected_token_value(
                        fluid_param, line, file_as_list, custom_message=f"Property {fluid_param} does \
                        not have a numerical value."))
            if nfo.check_token('DRYGAS_MFP', line):
                self.properties['DRYGAS_MFP'] = True

            # For EOS or compositional models, get required parameters
            if nfo.check_token('NHC', line):  # Get number of components
                self.eos_nhc = int(nfo.get_expected_token_value('NHC', line,
                                                                file_as_list,
                                                                custom_message="Property NHC does not \
                                                                have a numerical value."))
            if nfo.check_token('COMPONENTS', line):  # Get NHC components
                elems = line.split()
                components_index = elems.index('COMPONENTS')
                if self.eos_nhc and self.eos_nhc > 0:
                    self.eos_components = elems[components_index + 1:components_index + 1 + int(self.eos_nhc)]
            if nfo.check_token('TEMP', line):  # Get default EOS temperature
                self.eos_temp = float(nfo.get_expected_token_value(
                    'TEMP', line, file_as_list, custom_message="Property TEMP does not have a numerical value."))

            # Check for EOS options
            if nfo.check_token('EOSOPTIONS', line):
                if nfo.get_expected_token_value('EOSOPTIONS', line, file_as_list) in PVT_EOS_METHODS:
                    self.eos_options['EOS_METHOD'] = nfo.get_expected_token_value('EOSOPTIONS', line, file_as_list)
                else:
                    self.eos_options['EOS_METHOD'] = 'PR'
            # Find EOS single-word options, like CAPILLARYFLASH and add to list
            if [i for i in line.split() if i in PVT_EOSOPTIONS_PRIMARY_WORDS]:
                if 'EOS_OPT_PRIMARY_LIST' not in self.eos_options.keys():
                    self.eos_options['EOS_OPT_PRIMARY_LIST'] = []
                if not isinstance(self.eos_options['EOS_OPT_PRIMARY_LIST'], list):
                    raise ValueError(f"EOS_OPT_PRIMARY_LIST should be a list, instead \
                                     got {self.eos_options['EOS_OPT_PRIMARY_LIST']}")
                self.eos_options['EOS_OPT_PRIMARY_LIST'].extend([i for i in line.split() if i
                                                                 in PVT_EOSOPTIONS_PRIMARY_WORDS])
            # Find EOS key-value pairs, like LI_FACT 0.9 or FUGERR 5
            if [i for i in line.split() if i in PVT_EOSOPTIONS_PRIMARY_KEYS_FLOAT]:
                for key in PVT_EOSOPTIONS_PRIMARY_KEYS_FLOAT:
                    if nfo.check_token(key, line):
                        self.eos_options[key] = float(nfo.get_expected_token_value(key, line, file_as_list))
            if [i for i in line.split() if i in PVT_EOSOPTIONS_PRIMARY_KEYS_INT]:
                for key in PVT_EOSOPTIONS_PRIMARY_KEYS_INT:
                    if nfo.check_token(key, line):
                        self.eos_options[key] = int(nfo.get_expected_token_value(key, line, file_as_list))
            # Read TRANSITION, TRANS_TEST and PHASEID eos options, if present
            primary_keys2populate = ['TRANSITION', 'TRANS_TEST', 'PHASEID']
            primary_keys2populate_defaults = ['TEST', 'INCRP', '']
            secondary_keys = [PVT_EOSOPTIONS_TRANS_KEYS, PVT_EOSOPTIONS_TRANS_TEST_KEYS, PVT_EOSOPTIONS_PHASEID_KEYS]
            if [i for i in line.split() if i in primary_keys2populate]:
                trans_flag = True
            if trans_flag:
                for index in range(len(primary_keys2populate)):
                    pkey = primary_keys2populate[index]
                    p2key = primary_keys2populate_defaults[index]
                    sec_key = secondary_keys[index]
                    self.__populate_eos_opts_to_tertiary_keys(pkey, p2key, line, file_as_list, sec_key)
            # Read TRANS_OPTIMIZATION eos options, if present
            if nfo.check_token('TRANS_OPTIMIZATION', line):
                new_dict: dict[str, float] = {}
                for tert_key in PVT_EOSOPTIONS_TERTIARY_KEYS:
                    if nfo.check_token(tert_key, line):
                        potential_value = float(nfo.get_expected_token_value(tert_key, line, file_as_list))
                        if isinstance(potential_value, float):
                            new_dict[tert_key] = potential_value
                            self.eos_options['TRANS_OPTIMIZATION'] = new_dict

            # Identify beginning and ending line indices for different kinds tables in PVT file
            if start_reading_table:  # Figure out ending line indices for tables
                for table in PVT_TABLES_WITHOUT_ENDWORDS:
                    start_reading_table = self.__find_pvt_table_ending_index(table, line, pvt_table_indices,
                                                                             pvt_table_indices_dict,
                                                                             line_indx, table_being_read, False,
                                                                             start_reading_table)
                for table in PVT_TABLES_WITH_ENDWORDS:
                    start_reading_table = self.__find_pvt_table_ending_index(table, line, pvt_table_indices,
                                                                             pvt_table_indices_dict,
                                                                             line_indx, table_being_read, True,
                                                                             start_reading_table)
                for table in PVT_UNSAT_TABLE_INDICES:
                    start_reading_table = self.__find_pvt_table_ending_index(table, line, pvt_table_indices,
                                                                             pvt_table_indices_dict,
                                                                             line_indx, table_being_read, False,
                                                                             start_reading_table, pvt_unsat_keys)
            # Figure out beginning line indices for tables
            table_found = False
            for table in PVT_TABLE_KEYWORDS + PVT_UNSAT_TABLE_INDICES:
                if table in PVT_UNSAT_TABLE_INDICES:  # Work on undersaturated tables
                    new_line_indx = self.__find_pvt_table_starting_index(table, line, file_as_list, pvt_table_indices,
                                                                         pvt_table_indices_dict, table_being_read,
                                                                         line_indx, pvt_unsat_keys)
                else:  # All other tables
                    new_line_indx = self.__find_pvt_table_starting_index(table, line, file_as_list, pvt_table_indices,
                                                                         pvt_table_indices_dict, table_being_read,
                                                                         line_indx)
                if new_line_indx is not None:
                    line_indx = new_line_indx
                    table_found = True
                    break
            if table_found:
                continue
            # Check if this line represents the header of a PVT table
            table_header_row_options = ['PRES', 'DP', 'RV', 'INDEX', 'COMPONENT']
            header_row_flag = False
            for hr in table_header_row_options:
                if nfo.check_token(hr, line):
                    header_row_flag = True
            reading_a_table_flag = False
            for table_name in PVT_ALL_TABLE_KEYWORDS:
                if table_being_read[table_name]:
                    reading_a_table_flag = True
            if header_row_flag and reading_a_table_flag and not start_reading_table:
                start_reading_table = True  # If header row, start reading

            line_indx += 1

        # Read in tables
        for key in pvt_table_indices.keys():
            if key == 'IGS_CP':
                self.properties[key] = nfo.read_table_to_df(file_as_list[
                                                            pvt_table_indices[key][0]:pvt_table_indices[key][1]],
                                                            noheader=True)
            else:
                self.properties[key] = nfo.read_table_to_df(file_as_list[
                                                            pvt_table_indices[key][0]:pvt_table_indices[key][1]])
        for key in pvt_table_indices_dict.keys():
            self.properties[key] = {}
            for subkey in pvt_table_indices_dict[key].keys():
                property_key = self.properties[key]
                if not isinstance(property_key, dict):
                    raise ValueError(f"Property is not a dictionary: {str(self.properties[key])}")
                property_key[subkey] = nfo.read_table_to_df(file_as_list[
                                                            pvt_table_indices_dict[key][subkey][0]:
                                                            pvt_table_indices_dict[key][subkey][1]])



"""Class for collection of Nexus equilibration methods."""
from __future__ import annotations
from dataclasses import dataclass, field
import os
from typing import Optional, MutableMapping

from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.DataModels.FcsFile import FcsNexusFile
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.DataModels.NexusEquilMethod import NexusEquilMethod
from ResSimpy.DataModelBaseClasses.Equilibration import Equilibration
from ResSimpy.Utils.dynamic_method_manipulations import add_dynamic_method


@dataclass(kw_only=True)
class NexusEquilMethods(Equilibration):
    """Class for collection of Nexus equilibration methods.

    Attributes:
        inputs (dict[int, NexusEquilMethod]): Collection of Nexus equilibration methods, as a dictionary
        files (dict[int, NexusFile]): Dictionary collection of equilibration files, as defined in Nexus fcs file.
    """

    __inputs: MutableMapping[int, NexusEquilMethod]
    __files: dict[int, NexusFile]
    __properties_loaded: bool = False  # Used in lazy loading
    __model_unit_system: UnitSystem
    _model_files: Optional[FcsNexusFile] = field(default=None, repr=False, compare=False)

    def __init__(self, model_unit_system: UnitSystem, inputs: Optional[MutableMapping[int, NexusEquilMethod]] = None,
                 files: Optional[dict[int, NexusFile]] = None, assume_loaded: bool = False,
                 model_files: Optional[FcsNexusFile] = None) -> None:
        """Initialises the NexusEquilMethods class.

        Args:
            model_unit_system (UnitSystem): Unit system used in the model.
            inputs (Optional[MutableMapping[int, NexusEquilMethod]]): Collection of Nexus equilibration methods.
            files (Optional[dict[int, NexusFile]]): Collection of equilibration files, as defined in Nexus fcs file.
                Keyed by the method number.
            assume_loaded (bool): If True, assumes that the equilibration methods are already loaded.
            model_files (Optional[FcsNexusFile]): The FCS file that contains the model files.
        """
        if inputs:
            self.__inputs = inputs
        else:
            self.__inputs: MutableMapping[int, NexusEquilMethod] = {}
        if files:
            self.__files = files
        else:
            self.__files = {}
        self.__model_unit_system = model_unit_system
        self.__properties_loaded = assume_loaded
        self._model_files = model_files
        super().__init__()

    def __repr__(self) -> str:
        """Pretty printing equil methods."""
        if not self.__properties_loaded:
            self.load_equil_methods()
        printable_str = ''
        for table_num in self.__inputs.keys():
            printable_str += '\n--------------------------------\n'
            printable_str += f'EQUIL method {table_num}\n'
            printable_str += '--------------------------------\n'
            printable_str += self.__inputs[table_num].__repr__()
            printable_str += '\n'

        return printable_str

    @property
    def inputs(self) -> MutableMapping[int, NexusEquilMethod]:
        """Returns mapping of 'NexusEquilMethod' where keys are of type int.
        If properties are not loaded, it will load them.
        """
        if not self.__properties_loaded:
            self.load_equil_methods()
        return self.__inputs

    @property
    def files(self) -> dict[int, NexusFile]:
        """Returns dictionary of Nexus files where keys are of type int."""
        return self.__files

    def load_equil_methods(self) -> None:
        """Loads a collection of equilibration method files.
        This method checks if equil files are available and reads their properties into
        'NexusEquilmethod' objects.
        """
        # Read in equil properties from Nexus equil method files
        if self.__files is not None and len(self.__files) > 0:  # Check if equil files exist
            for table_num in self.__files.keys():  # For each equil property method
                equil_file = self.__files[table_num]
                if equil_file.location is None:
                    raise ValueError(f'Unable to find equil file: {equil_file}')
                if os.path.isfile(equil_file.location):
                    # Create NexusEquilMethod object
                    self.__inputs[table_num] = NexusEquilMethod(file=equil_file, input_number=table_num,
                                                                model_unit_system=self.__model_unit_system)
                    self.__inputs[table_num].read_properties()  # Populate object with equil properties in file
        self.__properties_loaded = True

    @property
    def model_unit_system(self) -> UnitSystem:
        """Return the model unit system."""
        return self.__model_unit_system

    @property
    def summary(self) -> str:
        """Returns a string summary of 'NexusEquilMethods' inputs in a dictionary."""
        equil_summary = ''
        for key, value in self.inputs.items():
            equil_summary += f'        {key}: {value.file.location}\n'

        return equil_summary

    @property
    def keyword(self) -> str:
        """Returns the keyword for the equilibration methods."""
        return 'EQUIL'

    def add_method(self, method: DynamicProperty, new_file_name: str, create_new_file: bool = False) -> None:
        """Adds a new equil method to the collection.

        Args:
            method (DynamicProperty): The equil method to add.
            new_file_name (str): The name of the file to save the method to.
            create_new_file (bool): Whether to create a new file for the method.
        """
        add_dynamic_method(dynamic_method_collection=self, method=method, new_file_name=new_file_name,
                           create_new_file=create_new_file)

    def _method_type(self) -> type[NexusEquilMethod]:
        """Returns the expected type of the dynamic property."""
        return NexusEquilMethod



"""Enum for defining frequency labels for outputs."""
from enum import Enum


class FrequencyEnum(Enum):
    """Enum for defining frequency labels for outputs."""

    UNDEFINED = 0
    NONE = 1
    DT = 2
    DTTOL = 3
    TIMES = 4
    FREQ = 5
    TNEXT = 6
    MONTHLY = 7
    QUARTERLY = 8
    YEARLY = 9
    TIMESTEP = 10
    TSTART = 11
    OFF = 12
    ON = 13



"""Abstract base class for a node in a network."""
from __future__ import annotations
from abc import ABC
from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.DataModelBaseClasses.NetworkObject import NetworkObject
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Units.AttributeMappings.NetworkUnitMapping import NetworkUnits


@dataclass(repr=False)
class Node(NetworkObject, ABC):
    type: Optional[str] = None
    depth: Optional[float] = None

    def __init__(self, properties_dict: Optional[DataObjectMixinDictType] = None,
                 date: Optional[str] = None, date_format: Optional[DateFormat] = None, start_date: Optional[str] = None,
                 unit_system: Optional[UnitSystem] = None, name: Optional[str] = None, type: Optional[str] = None,
                 depth: Optional[float] = None) -> None:
        """Initialises the NexusNode class.

        Args:
            properties_dict (Optional[dict]): dict of the properties to set on the object.
            date (Optional[str]): The date of the object.
            date_format (Optional[DateFormat]): The date format that the object uses.
            start_date (Optional[str]): The start date of the model. Required if the object uses a decimal TIME.
            unit_system (Optional[UnitSystem]): The unit system of the object e.g. ENGLISH, METRIC.
            name (Optional[str]): The name of the object.
            type (Optional[str]): Well type.
            depth (Optional[float]): The depth of the wellhead.
        """

        self.type = type
        self.depth = depth

        super().__init__(properties_dict=properties_dict, date=date, date_format=date_format, start_date=start_date,
                         unit_system=unit_system, name=name)

    @property
    def units(self) -> NetworkUnits:
        """Returns the attribute to unit map for the constraint."""
        return NetworkUnits(self.unit_system)



"""Class to hold Nexus Valve properties."""
from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Union
import numpy as np
import pandas as pd
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.NexusKeywords.valve_keywords import VALVE_TABLE_KEYWORDS, VALVE_RATE_KEYWORDS
from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.Units.AttributeMappings.DynamicPropertyUnitMapping import HydraulicsUnits
from ResSimpy.Enums.UnitsEnum import UnitSystem, TemperatureUnits
from ResSimpy.Utils.factory_methods import get_empty_dict_union
import ResSimpy.Nexus.nexus_file_operations as nfo


@dataclass(kw_only=True, repr=False)  # Doesn't need to write an _init_, _eq_ methods, etc.
class NexusValveMethod(DynamicProperty):
    """Class to hold Nexus Valve properties.

    Attributes:
        file (NexusFile): Nexus valve properties file object
        input_number (int): Valve properties method number in Nexus fcs file
        properties (dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                    dict[str, Union[float, pd.DataFrame]]]]):
            Dictionary holding all properties for a specific valve properties method. Defaults to empty dictionary.
    """

    # General parameters
    file: NexusFile
    properties: dict[str, Union[str, int, float, Enum, list[str], np.ndarray, pd.DataFrame,
                     dict[str, Union[float, pd.DataFrame]]]] = \
        field(default_factory=get_empty_dict_union)
    unit_system: UnitSystem
    __hash__ = DynamicProperty.__hash__

    def __init__(self, file: NexusFile, input_number: int, model_unit_system: UnitSystem,
                 properties: Optional[dict[str, Union[str, int, float, Enum, list[str], np.ndarray, pd.DataFrame,
                                      dict[str, Union[float, pd.DataFrame]]]]] = None) -> None:
        """Initialises the NexusValveMethod class.

        Args:
            file (NexusFile): NexusFile object associated with the valve method.
            input_number (int): method number for the valve method.
            model_unit_system (UnitSystem): unit system used in the model.
            properties: The properties found in the valve method.
        """
        if properties is not None:
            self.properties = properties
        else:
            self.properties = {}
        self.unit_system = model_unit_system
        super().__init__(input_number=input_number, file=file)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords: dict[str, tuple[str, type]] = {
            'VC': ('valve_coefficient', float),
        }
        return keywords

    @property
    def units(self) -> HydraulicsUnits:
        """Returns the attribute to unit map for the Valve method."""
        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']
        return HydraulicsUnits(unit_system=self.unit_system)

    def to_string(self) -> str:
        """Create string with valve data in Nexus file format."""
        printable_str = ''
        valve_dict = self.properties
        for key, value in valve_dict.items():
            if key == 'DESC' and isinstance(value, list):
                for desc_line in value:
                    printable_str += 'DESC ' + desc_line + '\n'
            elif isinstance(value, pd.DataFrame):
                table_text = f'{key}'
                if key in ['VALVE', 'ICV']:
                    for rate_key in VALVE_RATE_KEYWORDS:
                        if 'DP_RATE' in valve_dict.keys() and rate_key == valve_dict['DP_RATE']:
                            table_text += f' {rate_key}'
                printable_str += f'{table_text}\n'
                printable_str += value.to_string(na_rep='', index=False) + '\n'
                printable_str += 'END'+key+'\n'
                printable_str += '\n'
            elif isinstance(value, Enum):
                if isinstance(value, UnitSystem) or isinstance(value, TemperatureUnits):
                    printable_str += f'{value.value}\n'
            elif key not in ['DP_RATE']:
                printable_str += f'{key} {value}\n'
        return printable_str

    def read_properties(self) -> None:
        """Read Nexus valve file contents and populate the NexusValveMethod object."""
        file_as_list = self.file.get_flat_list_str_file

        # Check for common input data
        nfo.check_for_and_populate_common_input_data(file_as_list, self.properties)

        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']

        # Initialize flags and containers to use to record properties as we iterate through valve file contents
        # Dictionary to record start and ending indices for tables
        valve_table_indices: dict[str, list[int]] = {}
        # Dictionary of flags indicating which tables are being read
        table_being_read: dict[str, bool] = {}
        for table_name in VALVE_TABLE_KEYWORDS:
            table_being_read[table_name] = False

        line_indx = 0
        for line in file_as_list:

            # Find beginning and ending indices of tables
            for table_key in VALVE_TABLE_KEYWORDS:
                if nfo.check_token(table_key, line):
                    table_being_read[table_key] = True
                    valve_table_indices[table_key] = [line_indx + 1, len(file_as_list)]
                    # Find rate keyword if present
                    for rate_key in VALVE_RATE_KEYWORDS:
                        if nfo.check_token(rate_key, line):
                            self.properties['DP_RATE'] = rate_key  # Rate used to determine pressure drop across valve
                if table_being_read[table_key] and nfo.check_token('END' + table_key, line):
                    table_being_read[table_key] = False
                    valve_table_indices[table_key][1] = line_indx

            line_indx += 1

        # Read in tables
        for key in valve_table_indices.keys():
            self.properties[key] = nfo.read_table_to_df(file_as_list[
                                                        valve_table_indices[key][0]:valve_table_indices[key][1]])



"""Class for handling the set of Local Grid Refinements (LGR) in the NexusGrid."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from ResSimpy.GenericContainerClasses.LGRs import LGRs
from ResSimpy.Nexus.DataModels.StructuredGrid.NexusLGR import NexusLGR
from ResSimpy.FileOperations import file_operations as fo

if TYPE_CHECKING:
    from ResSimpy.Nexus.DataModels.StructuredGrid.NexusGrid import NexusGrid


@dataclass(kw_only=True)
class NexusLGRs(LGRs):
    """Class for handling the set of Local Grid Refinements (LGR) in the NexusGrid."""

    _lgr_list: list[NexusLGR] = field(default_factory=list)
    _grid_file_as_list: list[str] = field(default_factory=list)
    __has_been_loaded: bool = False
    __parent_grid: NexusGrid

    def __init__(self, parent_grid: NexusGrid, grid_file_as_list: None | list[str],
                 lgr_list: None | list[NexusLGR] = None, assume_loaded: bool = False) -> None:
        """Initializes the NexusLGRs class.

        Args:
            parent_grid (NexusGrid): The NexusGrid object that the LGRs belong to.
            grid_file_as_list (None | list[str]): List of strings representing the file to load the LGRs from.
            lgr_list (None | list[NexusLGR]): List of LGRs to initialize the class with. Defaults to None for loading
            purposes.
            assume_loaded (bool): Whether the LGRs have already been loaded. Defaults to False.
        """
        super().__init__(lgr_list=lgr_list)
        self._grid_file_as_list = grid_file_as_list if grid_file_as_list is not None else []
        self.__has_been_loaded = assume_loaded
        self.__parent_grid = parent_grid

    def load_lgrs(self) -> None:
        """Loads LGRs from a list of strings."""
        if self.__has_been_loaded:
            return
        # Implementation to load LGRs from the provided list
        start_cartref_index = -1
        end_cartref_index = -1
        lgr_name = ''
        for line_num, line in enumerate(self._grid_file_as_list):
            if fo.check_token('CARTREF', line):
                lgr_name = fo.get_expected_token_value('CARTREF',
                                                       token_line=line,
                                                       file_list=self._grid_file_as_list,
                                                       )
                start_cartref_index = line_num

            if fo.check_token('ENDREF', line):
                end_cartref_index = line_num

            if 0 < start_cartref_index < end_cartref_index:
                new_lgr = self.__read_lgr_table(
                    file_subsection=self._grid_file_as_list[start_cartref_index + 1:end_cartref_index],
                    lgr_name=lgr_name)
                self._lgr_list.append(new_lgr)
                start_cartref_index = -1
                end_cartref_index = -1

        self.__has_been_loaded = True

    def __read_lgr_table(self, file_subsection: list[str], lgr_name: str) -> NexusLGR:
        """Reads in a single LGR table. Returns a NexusLGR object from the read."""
        range_of_lgr = []

        filter_list = file_subsection.copy()
        for x in range(6):
            range_values = fo.get_expected_next_value(0, filter_list, filter_list[0], replace_with='')
            range_of_lgr.append(range_values)
        i1, i2, j1, j2, k1, k2 = (int(x) for x in range_of_lgr)

        refinement_lines: list[list[str]] = []
        for line in filter_list:
            if len(refinement_lines) == 3:
                # we expect 3 lines worth of refinement one for each
                break
            next_value = fo.get_next_value(0, [line])
            if fo.get_next_value(0, [line]) is None:
                continue
            line_values = []
            filtered_line = [line]
            while next_value is not None:
                next_value = fo.get_next_value(0, filtered_line, replace_with='')
                if next_value is not None:
                    line_values.append(next_value)

            refinement_lines.append(line_values)

        nx_ny_nz = []
        for ref_line in refinement_lines:
            processed_line = self.__expand_refinement_lines(ref_line)
            nx_ny_nz.append(processed_line)
        nx, ny, nz = nx_ny_nz

        return NexusLGR(parent_grid='ROOT', name=lgr_name, i1=i1, i2=i2, j1=j1, j2=j2, k1=k1, k2=k2,
                        nx=nx, ny=ny, nz=nz)

    @staticmethod
    def __expand_refinement_lines(refinement_line: list[str]) -> list[int]:
        """Expands the refinement lines into a list of integers."""
        store_final_values: list[int] = []
        for value in refinement_line:
            if '*' in value:
                number_values, val = value.split('*', 1)
                value_expanded = int(number_values) * [int(val)]
                store_final_values.extend(value_expanded)
            else:
                store_final_values.append(int(value))
        return store_final_values

    @property
    def lgr_list(self) -> list[NexusLGR]:
        """Collection of the LGRs in the NexusGrid."""
        if not self.__has_been_loaded:
            self.load_lgrs()
            self.__parent_grid.load_grid_properties_if_not_loaded()
        return self._lgr_list

    def get_all(self) -> list[NexusLGR]:
        """Returns the collection of LGRs in the NexusGrid."""
        return self.lgr_list

    def get(self, name: str) -> NexusLGR:
        """Returns the LGR with the specified name.

        Args:
            name (str): the name of the LGR to get from the grid.
        """
        lgr = next((x for x in self.lgr_list if x.name == name), None)
        if lgr is None:
            raise ValueError(f'No LGR with name {name} found in the NexusGrid.')
        return lgr



from dataclasses import dataclass
from typing import Mapping

from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.DataModelBaseClasses.DynamicPropertyContainer import DynamicPropertyContainer


@dataclass(kw_only=True)
class Hydraulics(DynamicPropertyContainer):
    """The abstract base class for a collection of hydraulics inputs.

    Attributes:
        inputs (dict[int, DynamicProperty]): Collection of hydraulics inputs, as a dictionary.
    """

    @property
    def summary(self) -> str:
        """Returns string summary of Hydraulics properties."""
        raise NotImplementedError("Implement this in a derived class")

    @property
    def inputs(self) -> Mapping[int, DynamicProperty]:
        """A Collection of hydraulics inputs, as a dictionary."""
        raise NotImplementedError("Implement this in the derived class")

    def add_method(self, method: DynamicProperty, new_file_name: str,
                   create_new_file: bool = False) -> None:
        """Adds a new hydraulics method to the collection.

        Args:
            method (Hydraulics): The hydraulics method to add.
            new_file_name (str): The name of the file to save the method to.
            create_new_file (bool): Whether to create a new file for the method.
        """
        raise NotImplementedError("Implement this in the derived class")



from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.DataModelBaseClasses.NetworkObject import NetworkObject
from ResSimpy.Nexus.NexusEnums.ActivationChangeEnum import ActivationChangeEnum
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Units.AttributeMappings.BaseUnitMapping import BaseUnitMapping


@dataclass(kw_only=True, repr=True)
class NexusAction(NetworkObject):
    """A class representing a single Nexus action."""

    action_time: float  # don't assume it is an int just yet
    change: ActivationChangeEnum
    connection: str

    def __init__(self, properties_dict: Optional[DataObjectMixinDictType] = None,
                 date: Optional[str] = None,
                 date_format: Optional[DateFormat] = None,
                 start_date: Optional[str] = None, unit_system: Optional[UnitSystem] = None) -> None:
        """Initializes the Nexus Action attributes."""

        # the fact that date is being explicitly set in the super call means that it would
        # technically be an unused parameter, but date is a requirement in the __init__ so we cannot
        # just get rid of it. So we implement a quick hack below.
        # This IF statement is necessary to pass the ruff check
        if date is None:
            pass

        # the action_time in the properties dict is the DATE
        action_time = ''  # expecting a float
        connection = ''

        if isinstance(properties_dict, dict):
            if isinstance(properties_dict['action_time'], float):
                action_time = str(properties_dict['action_time'])
            if isinstance(properties_dict['connection'], str):
                connection = properties_dict['connection']
        super().__init__(properties_dict=properties_dict, date=action_time,
                         date_format=date_format, start_date=start_date,
                         unit_system=unit_system, name=connection)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of keywords to attribute definitions."""
        keywords = {
            'ACTIONTIME': ('action_time', float),
            'ACTION': ('change', ActivationChangeEnum),
            'CONNECTION': ('connection', str),
        }
        return keywords

    @property
    def units(self) -> BaseUnitMapping:
        """Returns the attribute to unit map for the data object."""
        return BaseUnitMapping(unit_system=self.unit_system)



"""Class representing Drill Parameters in a Nexus Network."""
from __future__ import annotations
from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.DataModelBaseClasses.Drill import Drill
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat


@dataclass(kw_only=True)
class NexusDrill(Drill):
    """Class representing Drill Parameters in a Nexus Network.

    Attributes:
        drillsite (Optional[str]): The name of the site from which the well will be drilled.
        drill_time (Optional[int]): The amount of time before the rig is released.
        completion_time (Optional[int]): Time needed to complete the well.
        workover_time (Optional[int]): Amount of time needed to workover the well.
        rigs (Optional[str]): Rigs that can be used to drill the well.
        replacement(Optional[str]): Wells that could be used as a replacement for the well.
    """

    drill_site: Optional[str] = None
    drill_time: Optional[float] = None
    completion_time: Optional[float] = None
    workover_time: Optional[float] = None
    rigs: Optional[str] = None
    replacement: Optional[str] = None

    def __init__(self, unit_system: Optional[UnitSystem] = None, name: Optional[str] = None, date: Optional[str] = None,
                 date_format: Optional[DateFormat] = None, start_date: Optional[str] = None,
                 drillsite: Optional[str] = None, drill_time: Optional[float] = None,
                 completion_time: float = 0, workover_time: float = 0, rigs: str = 'ALL',
                 replacement: Optional[str] = None,
                 properties_dict: Optional[DataObjectMixinDictType] = None) -> None:
        """Initialises the NexusDrill class.

        Args:
            unit_system (Optional[UnitSystem]): The unit system of the object e.g. ENGLISH, METRIC.
            name (Optional[str]): The name of the well / will list the drill relates to.
            date (Optional[str]): The date of the object.
            date_format (Optional[DateFormat]): The date format that the object uses.
            start_date (Optional[str]): The start date of the model. Required if the object uses a decimal TIME.
            drillsite (Optional[str]): The name of the site from which the well will be drilled.
            drill_time (Optional[float]): The amount of time before the rig is released.
            completion_time (Optional[float]): Time needed to complete the well.
            workover_time (Optional[float]): Amount of time needed to workover the well.
            rigs (Optional[str]): Rigs that can be used to drill the well.
            replacement(Optional[str]): Wells that could be used as a replacement for the well.
            properties_dict (dict): dict of the properties to set on the object.
        """

        self.drill_site = drillsite
        self.drill_time = drill_time
        self.completion_time = completion_time
        self.workover_time = workover_time
        self.rigs = rigs
        self.replacement = replacement

        super().__init__(unit_system=unit_system, name=name, date_format=date_format, start_date=start_date, date=date,
                         properties_dict=properties_dict)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        nexus_mapping = {
            'WELL': ('name', str),
            'DRILLSITE': ('drill_site', str),
            'DRILLTIME': ('drill_time', float),
            'CMPLTIME': ('completion_time', float),
            'WORKTIME': ('workover_time', float),
            'RIGS': ('rigs', str),
            'REPLACEMENT': ('replacement', str),
        }

        return nexus_mapping

    @property
    def total_drill_time(self) -> float:
        """The total time to drill the well."""
        drill_time = self.drill_time if self.drill_time is not None else 0
        completion_time = self.completion_time if self.completion_time is not None else 0
        return drill_time + completion_time



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional
from uuid import UUID, uuid4

import numpy as np
import pandas as pd

from ResSimpy.FileOperations.File import File
from ResSimpy.FileOperations import file_operations as fo
from ResSimpy.Utils.grid_filtering_functions import grid_filter_file_as_list, filter_grid_array_definition
from ResSimpy.Utils.general_utilities import check_if_string_is_float


@dataclass
class GridArrayDefinition:
    """A class to define a grid array property.

    Args:
        name (str): the name of the grid array property.
        region_name (str): the name of the region this grid array property belongs to.
        modifier (Optional[str]): the modifier for the grid array property (e.g. CON, MULT, etc.)
        value (Optional[str]): the actual values for the grid array property in question. Can be an include file.
        mods (Optional[dict[str, pd.DataFrame]): if the grid array has an associated mod card we capture it.
        keyword_in_include_file (bool): an indicator to tell you if the grid array keyword was found in an inc file
        absolute_path (Optional[str]): the absolute path to the path if value is an include file.
    """
    name: str = field(default='', compare=False)
    region_name: str = field(default='', compare=False)
    modifier: Optional[str] = None
    value: Optional[str] = None
    mods: Optional[dict[str, pd.DataFrame]] = None
    keyword_in_include_file: bool = False
    absolute_path: Optional[str] = None
    array: Optional[np.ndarray] = None
    __id: UUID = field(default_factory=lambda: uuid4(), compare=False)

    def __init__(self, modifier: Optional[str] = None, value: Optional[str] = None,
                 mods: Optional[dict[str, pd.DataFrame]] = None, keyword_in_include_file: bool = False,
                 absolute_path: Optional[str] = None, array: Optional[np.ndarray] = None,
                 name: str = '', region_name: str = '') -> None:
        """Initializes a grid array definition object, representing a grid array property.

        Args:
            name (str): the name of the grid array property.
            region_name (str): the name of the region this grid array property belongs to.
            modifier (Optional[str]): the modifier for the grid array property (e.g. CON, MULT, etc.)
            value (Optional[str]): the actual values for the grid array property in question. Can be an include file.
            mods (Optional[dict[str, pd.DataFrame]): if the grid array has an associated mod card we capture it.
            keyword_in_include_file (bool): an indicator to tell you if the grid array keyword was found in an inc file
            absolute_path (Optional[str]): the absolute path to the path if value is an include file.
            array (Optional[np.ndarray]): The loaded array from the grid file. Loads from the absolute path.
        """
        self.__id = uuid4()
        self.name = name
        self.region_name = region_name
        self.modifier = modifier
        self.value = value
        self.mods = mods
        self.keyword_in_include_file = keyword_in_include_file
        self.absolute_path = absolute_path
        self.array = array

    def load_grid_array_definition_to_file_as_list(self) -> list[str]:
        """Loads the grid array definition to a file as a list of strings."""
        path = self.absolute_path
        if path is None and self.value is not None:
            sanitised_value = "".join(fo.strip_file_of_comments(self.value.splitlines()))
            # If the value is a path, load that file in. Otherwise, assume it is a grid array value.
            if sanitised_value.upper().isupper():  # Fastest way to check if there are any letters in a string
                path = self.value
            else:
                return sanitised_value.splitlines()
        elif path is None:
            raise FileNotFoundError('No file path found in the grid array definition')
        return fo.load_file_as_list(path)

    def get_array_from_file(self, x_range: None | int = None, y_range: None | int = None,
                            z_range: None | int = None) -> np.ndarray:
        """Returns a 1D numpy array from the grid array definition."""
        if self.array is not None:
            return self.array
        file_as_list = self.load_grid_array_definition_to_file_as_list()
        self.array = self.grid_file_as_list_to_numpy_array(file_as_list, x_range, y_range, z_range)
        return self.array

    def filtered_grid_array_def_as_file(self) -> File:
        """Filters the grid array definition."""
        return filter_grid_array_definition(self)

    @staticmethod
    def grid_file_as_list_to_numpy_array(file_as_list: list[str], x_range: None | int, y_range: None | int,
                                         z_range: None | int) -> np.ndarray:
        """Converts a list of strings to a numpy array."""
        # ensure the list of strings is filtered of comments and non-float values
        new_file_as_list = grid_filter_file_as_list(file_as_list)

        array_size = -1
        if x_range is not None and y_range is not None and z_range is not None:
            array_size = x_range * y_range * z_range

        grid_array = np.fromstring(' '.join(new_file_as_list), sep=' ', count=array_size)

        return grid_array

    def _get_array_or_value(self) -> np.ndarray | float:
        """Returns the array or value based on the modifier."""
        if self.modifier == 'VALUE':
            return self.get_array_from_file()
        elif self.modifier == 'CON':
            if self.value is None:
                raise ValueError('No value found for grid array definition.')
            return float(self.value)
        else:
            raise NotImplementedError('Only VALUE and CON are modifiers supported for grid array definitions.')

    def min(self) -> float:
        """Returns the minimum value of the grid array property."""
        array_or_value = self._get_array_or_value()
        if isinstance(array_or_value, (float, int)):
            return array_or_value
        else:
            return array_or_value.min()

    def max(self) -> float:
        """Returns the maximum value of the grid array property."""
        array_or_value = self._get_array_or_value()
        if isinstance(array_or_value, (float, int)):
            return array_or_value
        else:
            return array_or_value.max()

    def get_array(self) -> np.ndarray:
        """Returns the array from the grid array definition."""
        return self.get_array_from_file()

    @property
    def id(self) -> UUID:
        """Unique identifier for each object."""
        return self.__id

    def to_string(self) -> str:
        """Converts the object to a string."""
        if len(self.region_name) == 0:  # If no region name is provided
            grid_string = f'{self.name.upper()} {self.modifier}\n'
        else:
            grid_string = f'{self.name.upper()} {self.region_name} {self.modifier}\n'

        # Check if the value is an include file
        if (self.value is not None and isinstance(self.value, str)
                and not check_if_string_is_float(self.value.split()[0])):
            grid_string += f'INCLUDE {self.value}\n'
        else:  # If value is not an include file, we assume a string of values provided
            if self.modifier in ['CON', 'VALUE', 'XVAR', 'YVAR', 'ZVAR', 'LAYER', 'DIP', 'MULT']:
                grid_string += f'{self.value}\n'
            elif self.modifier == 'NONE':
                grid_string += ''

        # Add the mods if they exist
        if self.mods is not None:
            for mod_name, mod_df in self.mods.items():
                grid_string += f'{mod_name.upper()}\n'
                grid_string += mod_df.to_string(index=False, header=False) + '\n'

        return grid_string



"""Functions for loading a well list from a surface network file."""
from typing import TypeVar, Type

from ResSimpy.DataModelBaseClasses.NetworkList import NetworkList
import ResSimpy.Nexus.nexus_file_operations as nfo
from ResSimpy.Enums.HowEnum import OperationEnum
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat

SubTypeNetworkList = TypeVar('SubTypeNetworkList', bound=NetworkList)


def load_table_to_lists(file_as_list: list[str], row_object: Type[SubTypeNetworkList], table_header: str,
                        current_date: None | str,
                        date_format: DateFormat, previous_lists: None | list[SubTypeNetworkList] = None,
                        table_start_index: int = 0) -> list[tuple[SubTypeNetworkList, int]]:
    """Loads a well list from a surface network file.

    Args:
        file_as_list (list[str]): The surface network file as a list of strings.
        row_object (NetworkList): The object to be loaded.
        table_header (str): The header of the table to read from.
        current_date (None | str): The current date of the simulation.
        date_format (Optional[DateFormat]): The date format of the objects being loaded.
        previous_lists (None | list[tuple[NexusWellList, int]]): /
        The list of tuples of already loaded in NetworkLists with their starting index. Defaults to None.
        table_start_index (int): The starting index of the table. Defaults to 0.

    Returns:
        NexusWellList: A NexusWellList object containing the well list data.
    """
    if previous_lists is None:
        previous_lists = []
    if current_date is None:
        current_date = ''
    list_name = ''
    for line in file_as_list:
        if nfo.check_token(table_header, line):
            list_name = nfo.get_expected_token_value(table_header, line, file_as_list)

    previous_list_object = get_previous_list(list_name=list_name, lists=previous_lists)

    new_list_object = load_list_from_table(table_as_list_str=file_as_list, current_date=current_date,
                                           previous_list_object=previous_list_object, list_name=list_name,
                                           date_format=date_format, row_object=row_object, table_header=table_header)

    return [(new_list_object, table_start_index)]


def get_previous_list(list_name: str, lists: list[SubTypeNetworkList]) -> SubTypeNetworkList | None:
    """Gets the previous network list from a list of well lists.

    Args:
        list_name (str): The name of the well list to get.
        lists (list[NetworkList]): The list of well lists to search.

    Returns:
        NetworkList | None: The previous well list if it exists, None otherwise.
    """

    # Find the latest matching welllist
    matching_lists = [x for x in lists if x.name == list_name]
    sorted_lists = sorted(matching_lists, key=lambda x: x.iso_date)
    return None if len(sorted_lists) == 0 else sorted_lists[-1]


def load_list_from_table(table_as_list_str: list[str], row_object: Type[SubTypeNetworkList], table_header: str,
                         current_date: str,
                         list_name: str, date_format: DateFormat,
                         previous_list_object: SubTypeNetworkList | None = None) -> SubTypeNetworkList:
    """Loads a well list from a table taken from a surface network file.

    Currently does not support Wildcards in welllists.

    Args:
        table_as_list_str (list[str]): The subsection of surface network file as a list of strings.
        row_object (NetworkList): The object to be loaded.
        table_header (str): The header of the table to read from.
        current_date (str): The current date of the simulation.
        list_name (str): The name of the well list.
        date_format (Optional[DateFormat]): The date format of the Welllist being loaded.
        previous_list_object (NexusWellList | None): The previous well list. Defaults to None.

    Returns:
        NexusWellList: A NexusWellList object containing the well list data.
    """
    store_list = previous_list_object.elements_in_the_list.copy() if previous_list_object is not None else []
    operation: OperationEnum | None = None
    for line in table_as_list_str:
        if nfo.check_token('CLEAR', line):
            # clear the existing welllist
            store_list = []
            continue

        if nfo.check_token('ADD', line):
            operation = OperationEnum.ADD
            continue
        elif nfo.check_token('REMOVE', line):
            operation = OperationEnum.REMOVE
            continue

        if operation is None:
            continue
        if nfo.check_token('END' + table_header, line):
            break

        # otherwise the well should be removed/added from/to the welllist.
        name = nfo.get_next_value(0, [line])
        if name is None:
            # handle the empty line case
            continue

        if operation == OperationEnum.ADD:
            store_list.append(name)
        elif operation == OperationEnum.REMOVE:
            if name in store_list:
                store_list.remove(name)

    # TODO may also need to store the line number for the loaded well list
    return row_object(name=list_name, elements_in_the_list=store_list, date=current_date,
                      date_format=date_format)



from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataModelBaseClasses.NetworkObject import NetworkObject
from ResSimpy.Nexus.NexusEnums.ActivationChangeEnum import ActivationChangeEnum
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Units.AttributeMappings.BaseUnitMapping import BaseUnitMapping


@dataclass(kw_only=True, repr=True)
class NexusActivationChange(NetworkObject):
    change: ActivationChangeEnum
    is_constraint_change: bool

    def __init__(self, change: ActivationChangeEnum, date: Optional[str] = None,
                 date_format: Optional[DateFormat] = None, start_date: Optional[str] = None,
                 name: Optional[str] = None, is_constraint_change: bool = False) -> None:
        """Initialises the NexusTarget class.

        Args:
            change: ActivationChangeEnum: The change in activation (i.e. whether it has been activated or deactivated).
            date (Optional[str]): The date of the object.
            date_format (Optional[DateFormat]): The date format that the object uses.
            start_date (Optional[str]): The start date of the model. Required if the object uses a decimal TIME.
            name (Optional[str]): The name of the object that is being activated / deactivated.
            is_constraint_change (bool): Whether the activation change is the result of a constraint block or not.
        """
        self.change = change
        self.is_constraint_change = is_constraint_change
        super().__init__(date_format=date_format, start_date=start_date, name=name, date=date)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of keywords to attribute definitions."""
        raise NotImplementedError("Not relevant for this class")

    @property
    def units(self) -> BaseUnitMapping:
        """Returns the attribute to unit map for the data object."""
        raise NotImplementedError("Not relevant for this class")



"""The abstract base class for all simulators."""
from __future__ import annotations
import os
from abc import ABC, abstractmethod
from dataclasses import dataclass

from ResSimpy.DataModelBaseClasses.Aquifer import Aquifer
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.DataModelBaseClasses.Equilibration import Equilibration
from ResSimpy.FileOperations.File import File
from ResSimpy.DataModelBaseClasses.Gaslift import Gaslift
from ResSimpy.DataModelBaseClasses.Grid import Grid
from ResSimpy.GenericContainerClasses.Hydraulics import Hydraulics
from ResSimpy.DataModelBaseClasses.Network import Network
from ResSimpy.DataModelBaseClasses.PVT import PVT
from ResSimpy.DataModelBaseClasses.RelPerm import RelPerm
from ResSimpy.DataModelBaseClasses.Reporting import Reporting
from ResSimpy.DataModelBaseClasses.Rock import Rock
from ResSimpy.DataModelBaseClasses.Separator import Separator
from ResSimpy.DataModelBaseClasses.Valve import Valve
from ResSimpy.DataModelBaseClasses.Water import Water
from ResSimpy.GenericContainerClasses.Wells import Wells
from ResSimpy.Time.ISODateTime import ISODateTime


@dataclass(kw_only=True, init=False)
class Simulator(ABC):
    _start_date: str
    _origin: str
    _wells: Wells
    _pvt: PVT
    _separator: Separator
    _water: Water
    _equil: Equilibration
    _rock: Rock
    _relperm: RelPerm
    _valve: Valve
    _aquifer: Aquifer
    _hydraulics: Hydraulics
    _gaslift: Gaslift
    _network: Network
    _grid: None | Grid
    _model_files: File
    _default_units: UnitSystem
    _reporting: Reporting

    def __repr__(self) -> str:
        """Pretty printing Simulator data."""
        printable_str = f'Origin: {self.origin}\n'
        printable_str += f'Full path: {self.model_files.location}\n'
        printable_str += f'Start date: {self.start_date}\n'
        printable_str += f'Date format: {self.get_date_format()}\n'
        printable_str += f'Default units: {str(self.default_units)}\n'
        return printable_str

    """Class Properties"""

    @property
    def start_date(self) -> str:
        """Returns the start date of the model as a string."""
        return self._start_date

    @start_date.setter
    def start_date(self, value: str) -> None:
        self._start_date = value

    @property
    def wells(self) -> Wells:
        """Returns an instance of wells."""
        return self._wells

    @property
    def pvt(self) -> PVT:
        """Returns inputs of pvt (production validation testing)."""
        return self._pvt

    @property
    def separator(self) -> Separator:
        """Returns an instance of separator property."""
        return self._separator

    @property
    def water(self) -> Water:
        """Returns an instance of water property."""
        return self._water

    @property
    def equil(self) -> Equilibration:
        """Returns an instance of equilibration."""
        return self._equil

    @property
    def rock(self) -> Rock:
        """Returns an instance of rock property."""
        return self._rock

    @property
    def relperm(self) -> RelPerm:
        """Returns an instance of relative permeability."""
        return self._relperm

    @property
    def valve(self) -> Valve:
        """Returns an instance of valve property."""
        return self._valve

    @property
    def aquifer(self) -> Aquifer:
        """Returns an instance of aquifer."""
        return self._aquifer

    @property
    def hydraulics(self) -> Hydraulics:
        """Returns an instance of hydraulics."""
        return self._hydraulics

    @property
    def gaslift(self) -> Gaslift:
        """Returns an instance of gas lift property."""
        return self._gaslift

    @property
    def network(self) -> Network:
        """Returns an instance of network."""
        return self._network

    @property
    def default_units(self) -> UnitSystem:
        """Returns the default units to use if no units are found."""
        return self._default_units

    @staticmethod
    @abstractmethod
    def sim_default_unit_system() -> UnitSystem:
        """Returns the default unit system used by the Simulator."""
        raise NotImplementedError("This method has not been implemented for this simulator yet")

    @property
    def grid(self) -> None | Grid:
        """Pass the grid information to the front end."""
        return self._grid

    @property
    def origin(self) -> str:
        """Returns an instance of origin as a string."""
        return self._origin

    @origin.setter
    def origin(self, value: None | str) -> None:
        if value is None:
            raise ValueError(f'Origin path to model is required. Instead got {value}.')
        self._origin: str = value.strip()

    @property
    def model_location(self) -> str:
        """Returns the location of the model."""
        return os.path.dirname(self._origin)

    @property
    def model_files(self) -> File:
        """Returns model file of this instance."""

        return self._model_files

    @property
    @abstractmethod
    def start_iso_date(self) -> ISODateTime:
        """Returns the start date of the model in ISO format."""
        raise NotImplementedError("Implement this method on the derived class")

    """ Class Methods """

    @staticmethod
    @abstractmethod
    def get_fluid_type(surface_file_content: list[str]) -> str:
        """Returns the fluid type.

        Args:
            surface_file_content(list[str]): Reads the file content of surface with each line as a new entry in the list
        """
        raise NotImplementedError("This method has not been implemented for this simulator yet")

    @abstractmethod
    def set_output_path(self, path: str) -> None:
        """Initialise the output to the declared path (output location).

        Args:
            path(str): Declared location of the path.
        """
        raise NotImplementedError("Implement this method on the derived class")

    @abstractmethod
    def get_date_format(self) -> str:
        """Returns date format as a string."""
        raise NotImplementedError("Implement this method on the derived class")

    @abstractmethod
    def write_out_new_model(self, new_location: str, new_model_name: str) -> None:
        """Writes out a new version of the model to the location supplied.

        Args:
        new_location (str): Path to write the contents of the model to.
        new_model_name (str): The name for the model that will be created.
        """
        raise NotImplementedError("Implement this method on the derived class")



"""Class to represent Nexus grid array functions."""
from dataclasses import dataclass
from typing import Optional
import pandas as pd

from ResSimpy.DataModelBaseClasses.GridArrayFunction import GridArrayFunction


@dataclass
class NexusGridArrayFunction(GridArrayFunction):
    """Class to hold Nexus grid array function attributes.

    Attributes:
        function_table (pd.DataFrame): Dependency between input and output grid arrays, represented in tabular form
        function_table_m (int): Maximum number of function table entries which can be used in tabular interpolation
        function_table_p_list (list[float]): Shifts of the first, second, ..., and N-th input arrays
    """

    function_table: Optional[pd.DataFrame] = None
    function_table_m: Optional[int] = None
    function_table_p_list: Optional[list[float]] = None

    def to_string(self) -> str:
        """Creates a string representation of the object."""

        # compile the function string
        func_string = "FUNCTION"

        # create beginning of string by checking region type
        if self.region_number is not None and self.region_type is not None:
            # the region number immediately follows the region type
            reg_num_str = ' '.join(str(num) for num in self.region_number)
            func_string += f" {self.region_type}\n{reg_num_str}\n"
        elif self.region_type is None:
            # if there is no region type, we just write FUNCTION
            func_string += "\n"

        if self.blocks is not None:
            # after region type comes BLOCKS
            block_num_str = ' '.join(str(num) for num in self.blocks)
            func_string += f"BLOCKS {block_num_str}\n"

        if self.grid_name is not None:
            # after blocks comes the grid
            func_string += f'GRID {self.grid_name}\n'

        if self.input_range is not None:
            input_vals = (
                ' '.join(f'{i:.5e}' if len(str(i)) > 5 else str(i) for pair in self.input_range for i in pair))
            func_string += 'RANGE INPUT' + ' ' + input_vals + '\n'

        if self.output_range is not None:
            output_vals = (
                ' '.join(f'{i:.5e}' if len(str(i)) > 5 else str(i) for pair in self.output_range for i in pair))
            func_string += 'RANGE OUTPUT' + ' ' + output_vals + '\n'

        if self.drange is not None:
            drange_vals = ' '.join(f'{num:.5e}' if len(str(num)) > 5 else str(num) for num in self.drange)
            func_string += 'DRANGE' + ' ' + drange_vals + '\n'

        if self.function_type is not None and self.function_type.value is not None:
            # ANALYT must always precede function type
            func_vals = ''
            if self.function_values is not None:
                func_vals = ' '+' '.join(str(num) for num in self.function_values)
            func_string += f'ANALYT {self.function_type.value}' + func_vals + '\n'

        # specify the end of the string with output/input arrays
        input_arr = output_arr = ''
        if self.input_array is not None and self.output_array is not None:
            input_arr = ' '.join(num for num in self.input_array)
            output_arr = ' '.join(num for num in self.output_array)

        return func_string + input_arr + ' OUTPUT ' + output_arr + '\n'

    def __str__(self) -> str:
        return self.to_string()

    def __repr__(self) -> str:

        beginning = 'NexusGridArrayFunction('
        close_it = ')'

        # for some reason python does not pick up dataclass_fields, but it still works.
        running_str = ''
        for param in self.__dataclass_fields__.values():

            field_name = param.name
            class_attr = getattr(self, param.name)

            if class_attr is not None:
                # covert class_attr to str since there are some lists
                running_str += field_name + '=' + str(class_attr) + ','

        # strings are immutable in python so need to overwrite with slice
        # this takes care of the extra ',' caused at the end of the above loop
        running_str = running_str[0:-1]
        return beginning + running_str + close_it



"""File object for handling FCS top level files for Nexus."""
from __future__ import annotations
from dataclasses import dataclass, field
import os
import warnings

from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.FileOperations.File import File
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from typing import Optional, Generator

# Use correct Self type depending upon Python version
import sys

from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from ResSimpy.Utils.factory_methods import get_empty_dict_int_nexus_file, get_empty_list_str, get_empty_list_file
from ResSimpy.Nexus.NexusKeywords.fcs_keywords import FCS_KEYWORDS
import ResSimpy.Nexus.nexus_file_operations as nfo
import ResSimpy.FileOperations.file_operations as fo
from ResSimpy.Utils.generic_repr import generic_str
from datetime import datetime


@dataclass(kw_only=True)
class FcsNexusFile(NexusFile):
    restart_file: Optional[NexusFile] = None
    structured_grid_file: Optional[NexusFile] = None
    options_file: Optional[NexusFile] = None
    runcontrol_file: Optional[NexusFile] = None
    override_file: Optional[NexusFile] = None
    eos_default_file: Optional[NexusFile] = None
    well_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    surface_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    rock_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    relperm_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    pvt_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    water_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    equil_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    tracer_init_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    aquifer_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    hyd_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    valve_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    separator_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    ipr_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    gaslift_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    pump_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    compressor_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    choke_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    icd_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    esp_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    polymer_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    adsorption_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    flux_in_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    files_info: list[tuple[Optional[str], Optional[str], Optional[datetime]]]
    multi_reservoir_files: Optional[dict[str, FcsNexusFile]] = field(default_factory=dict)

    def __init__(
            self, location: str,
            include_locations: Optional[list[str]] = None,
            origin: Optional[str] = None,
            include_objects: Optional[list[NexusFile]] = None,
            file_content_as_list: Optional[list[str]] = None,
            restart_file: Optional[NexusFile] = None,
            structured_grid_file: Optional[NexusFile] = None,
            options_file: Optional[NexusFile] = None,
            runcontrol_file: Optional[NexusFile] = None,
            override_file: Optional[NexusFile] = None,
            eos_default_file: Optional[NexusFile] = None,
            well_files: Optional[dict[int, NexusFile]] = None,
            surface_files: Optional[dict[int, NexusFile]] = None,
            rock_files: Optional[dict[int, NexusFile]] = None,
            relperm_files: Optional[dict[int, NexusFile]] = None,
            pvt_files: Optional[dict[int, NexusFile]] = None,
            water_files: Optional[dict[int, NexusFile]] = None,
            equil_files: Optional[dict[int, NexusFile]] = None,
            tracer_init_files: Optional[dict[int, NexusFile]] = None,
            aquifer_files: Optional[dict[int, NexusFile]] = None,
            hyd_files: Optional[dict[int, NexusFile]] = None,
            valve_files: Optional[dict[int, NexusFile]] = None,
            separator_files: Optional[dict[int, NexusFile]] = None,
            ipr_files: Optional[dict[int, NexusFile]] = None,
            gaslift_files: Optional[dict[int, NexusFile]] = None,
            pump_files: Optional[dict[int, NexusFile]] = None,
            compressor_files: Optional[dict[int, NexusFile]] = None,
            choke_files: Optional[dict[int, NexusFile]] = None,
            icd_files: Optional[dict[int, NexusFile]] = None,
            esp_files: Optional[dict[int, NexusFile]] = None,
            polymer_files: Optional[dict[int, NexusFile]] = None,
            adsorption_files: Optional[dict[int, NexusFile]] = None,
            flux_in_files: Optional[dict[int, NexusFile]] = None,
            multi_reservoir_files: Optional[dict[str, FcsNexusFile]] = None,
    ) -> None:
        """Initialises the FcsNexusFile class.

        Args:
            location: str: The file path to the fcs file.
            include_locations: Optional[list[str]]: list of file paths to the included files.
            origin: Optional[str]: The file path to the file that included this file. None for top level files.
            include_objects: Optional[list[NexusFile]: list of NexusFile objects that are included in this file.
            file_content_as_list: Optional[list[str]]: list of strings representing the content of the file.
            restart_file: Optional[NexusFile]: The restart file for the fcs file.
            structured_grid_file: Optional[NexusFile]: The structured grid file for the fcs file.
            options_file: Optional[NexusFile]: The options file for the fcs file.
            runcontrol_file: Optional[NexusFile]: The runcontrol file for the fcs file.
            override_file: Optional[NexusFile]: The override file for the fcs file.
            eos_default_file: Optional[NexusFile]: The eos default file for the fcs file.
            well_files: Optional[dict[int, NexusFile]: Collection of well files for the fcs file. Indexed by method
            number.
            surface_files: Optional[dict[int, NexusFile]: Collection of surface files for the fcs file. Indexed by
            method number.
            rock_files: Optional[dict[int, NexusFile]: Collection of rock files for the fcs file. Indexed by method
            number.
            relperm_files: Optional[dict[int, NexusFile]: Collection of relperm files for the fcs file. Indexed by
            method number.
            pvt_files: Optional[dict[int, NexusFile]: Collection of pvt files for the fcs file. Indexed by method
            number.
            water_files: Optional[dict[int, NexusFile]: Collection of water files for the fcs file. Indexed by method
            number.
            equil_files: Optional[dict[int, NexusFile]: Collection of equil files for the fcs file. Indexed by method
            number.
            tracer_init_files: Optional[dict[int, NexusFile]: Collection of tracer init files for the fcs file. Indexed
            by method number.
            aquifer_files: Optional[dict[int, NexusFile]: Collection of aquifer files for the fcs file. Indexed by
            method number.
            hyd_files: Optional[dict[int, NexusFile]: Collection of hyd files for the fcs file. Indexed by method
            number.
            valve_files: Optional[dict[int, NexusFile]: Collection of valve files for the fcs file. Indexed by method
            number.
            separator_files: Optional[dict[int, NexusFile]: Collection of separator files for the fcs file. Indexed by
            method number.
            ipr_files: Optional[dict[int, NexusFile]: Collection of ipr files for the fcs file. Indexed by method
            number.
            gaslift_files: Optional[dict[int, NexusFile]: Collection of gas lift files for the fcs file. Indexed by
            method number.
            pump_files: Optional[dict[int, NexusFile]: Collection of pump files for the fcs file. Indexed by method
            number.
            compressor_files: Optional[dict[int, NexusFile]: Collection of compressor files for the fcs file. Indexed by
            method number.
            choke_files: Optional[dict[int, NexusFile]: Collection of choke files for the fcs file. Indexed by method
            number.
            icd_files: Optional[dict[int, NexusFile]: Collection of icd files for the fcs file. Indexed by method
            number.
            esp_files: Optional[dict[int, NexusFile]: Collection of esp files for the fcs file. Indexed by method
            number.
            polymer_files: Optional[dict[int, NexusFile]: Collection of polymer files for the fcs file. Indexed by
            method number.
            adsorption_files: Optional[dict[int, NexusFile]: Collection of adsorption files for the fcs file. Indexed by
            method number.
            flux_in_files: Optional[dict[int, NexusFile]: Collection of flux in files for the fcs file. Indexed by
            method number.
            multi_reservoir_files: Optional[dict[str, FcsNexusFile]]: Collection of multi-reservoir files for the
            top level fcs file. Indexed by reservoir name.
        """
        self.restart_file = restart_file

        self.structured_grid_file = structured_grid_file
        self.options_file = options_file
        self.runcontrol_file = runcontrol_file
        self.override_file = override_file
        self.eos_default_file = eos_default_file
        self.surface_files = surface_files
        self.well_files = well_files if well_files is not None else get_empty_dict_int_nexus_file()
        self.rock_files = rock_files if rock_files is not None else get_empty_dict_int_nexus_file()
        self.relperm_files = relperm_files if relperm_files is not None else get_empty_dict_int_nexus_file()
        self.pvt_files = pvt_files if pvt_files is not None else get_empty_dict_int_nexus_file()
        self.water_files = water_files if water_files is not None else get_empty_dict_int_nexus_file()
        self.equil_files = equil_files if equil_files is not None else get_empty_dict_int_nexus_file()
        self.tracer_init_files = tracer_init_files if tracer_init_files is not None else get_empty_dict_int_nexus_file()
        self.aquifer_files = aquifer_files if aquifer_files is not None else get_empty_dict_int_nexus_file()
        self.hyd_files = hyd_files if hyd_files is not None else get_empty_dict_int_nexus_file()
        self.valve_files = valve_files if valve_files is not None else get_empty_dict_int_nexus_file()
        self.separator_files = separator_files if separator_files is not None else get_empty_dict_int_nexus_file()
        self.ipr_files = ipr_files if ipr_files is not None else get_empty_dict_int_nexus_file()
        self.gaslift_files = gaslift_files if gaslift_files is not None else get_empty_dict_int_nexus_file()
        self.pump_files = pump_files if pump_files is not None else get_empty_dict_int_nexus_file()
        self.compressor_files = compressor_files if compressor_files is not None else get_empty_dict_int_nexus_file()
        self.choke_files = choke_files if choke_files is not None else get_empty_dict_int_nexus_file()
        self.icd_files = icd_files if icd_files is not None else get_empty_dict_int_nexus_file()
        self.esp_files = esp_files if esp_files is not None else get_empty_dict_int_nexus_file()
        self.polymer_files = polymer_files if polymer_files is not None else get_empty_dict_int_nexus_file()
        self.adsorption_files = adsorption_files if adsorption_files is not None else get_empty_dict_int_nexus_file()
        self.flux_in_files = flux_in_files if flux_in_files is not None else get_empty_dict_int_nexus_file()
        self.multi_reservoir_files: dict[str, FcsNexusFile] = multi_reservoir_files if (multi_reservoir_files
                                                                                        is not None) else {}
        self.files_info = []
        super().__init__(location=location, include_locations=include_locations, origin=origin,
                         include_objects=include_objects, file_content_as_list=file_content_as_list)

    def __repr__(self) -> str:
        printable_str = f'fcs file location: {self.location}\n'
        printable_str += '\tFCS file contains:\n'
        for file_type in self.fcs_keyword_map_single().values():
            if getattr(self, file_type, None) is not None:
                printable_str += f'\t\t{file_type}: {getattr(self, file_type).location}\n'
        for multi_file_type in self.fcs_keyword_map_multi().values():
            multi_file_list = getattr(self, multi_file_type, None)
            if multi_file_list is not None and len(multi_file_list) > 0:
                printable_str += f'\t\t{multi_file_type}: {len(multi_file_list)}\n'
        return printable_str

    def __str__(self) -> str:
        return generic_str(self)

    @classmethod
    def generate_fcs_structure(cls: type[Self], fcs_file_path: str, recursive: bool = True) -> Self:
        """Creates an instance of the FcsNexusFile, populates it through looking through the different keywords \
            in the FCS and assigning the paths to objects.

        Args:
        ----
            fcs_file_path (str): path to the fcs file of interest
            recursive (bool, optional): Whether the NexusFile structure will be recursively created. Defaults to True.

        Raises:
        ------
            FileNotFoundError: if the fcs file cannot be found
            ValueError: if no content can be found within the fcsfile

        Returns:
        -------
            FcsNexusFile: instance of a FcsNexusFile for a given fcs file path
        """
        fcs_file = cls(location=fcs_file_path)
        fcs_file.include_objects = get_empty_list_file()
        fcs_file.file_content_as_list = get_empty_list_str()
        fcs_file.include_locations = get_empty_list_str()

        # guard against bad links/empty files:
        if not os.path.isfile(fcs_file_path):
            raise FileNotFoundError(f'fcs file not found for path {fcs_file_path}')
        origin_path = fcs_file_path
        fcs_nexus_file = NexusFile.generate_file_include_structure(simulator_type=NexusFile, file_path=fcs_file_path,
                                                                   origin=None)
        fcs_file.files_info.append((fcs_nexus_file.location, fcs_nexus_file.linked_user,
                                    fcs_nexus_file.last_modified))

        fcs_file.linked_user = fcs_nexus_file.linked_user
        fcs_file.last_modified = fcs_nexus_file.last_modified

        flat_fcs_file_content = fcs_nexus_file.get_flat_list_str_file
        if flat_fcs_file_content is None or fcs_file.file_content_as_list is None:
            raise ValueError(f'FCS file not found, no content for {fcs_file_path=}')
        fcs_file.file_content_as_list = flat_fcs_file_content
        for i, line in enumerate(flat_fcs_file_content):
            if not nfo.nexus_token_found(line, valid_list=FCS_KEYWORDS):
                continue
            key = nfo.get_next_value(start_line_index=i, file_as_list=flat_fcs_file_content, search_string=line)
            if key is None:
                warnings.warn(f'get next value failed to find a suitable token in {line}')
                continue
            key = key.upper()
            value = fo.get_token_value(key, line, flat_fcs_file_content[i::])
            if value is None:
                warnings.warn(f'No value found for {key}, skipping file')
                continue
            # TODO handle methods / sets instead of getting full file path
            if key in cls.fcs_keyword_map_multi():
                # for keywords that have multiple methods we store the value in a dictionary
                # with the method number and the NexusFile object
                _, method_string, method_number, value = (
                    fo.get_multiple_expected_sequential_values(flat_fcs_file_content[i:], 4, ['NORPT'])
                )
                full_file_path = fo.get_full_file_path(value, origin_path)
                nexus_file = NexusFile.generate_file_include_structure(simulator_type=NexusFile, file_path=value,
                                                                       origin=fcs_file_path, recursive=recursive,
                                                                       top_level_file=True)
                fcs_property = getattr(fcs_file, cls.fcs_keyword_map_multi()[key])
                # manually initialise if the property is still a None after class instantiation
                if fcs_property is None:
                    fcs_property = get_empty_dict_int_nexus_file()
                # shallow copy to maintain list elements pointing to nexus_file that are
                # stored in the file_content_as_list
                fcs_property_list = fcs_property.copy()
                fcs_property_list.update({int(method_number): nexus_file})
                # set the attribute in the FcsNexusFile instance
                setattr(fcs_file, cls.fcs_keyword_map_multi()[key], fcs_property_list)
                fcs_file.include_objects.append(nexus_file)
                fcs_file.include_locations.append(full_file_path)
                fcs_file.files_info.append((nexus_file.location, nexus_file.linked_user,
                                            nexus_file.last_modified))
            elif key in cls.fcs_keyword_map_single():
                full_file_path = fo.get_full_file_path(value, origin_path)
                skip_arrays = True if key == 'STRUCTURED_GRID' else False
                nexus_file = NexusFile.generate_file_include_structure(simulator_type=NexusFile, file_path=value,
                                                                       origin=fcs_file_path, recursive=recursive,
                                                                       top_level_file=True, skip_arrays=skip_arrays)
                setattr(fcs_file, cls.fcs_keyword_map_single()[key], nexus_file)
                fcs_file.include_objects.append(nexus_file)
                fcs_file.include_locations.append(full_file_path)
                fcs_file.files_info.append((nexus_file.location, nexus_file.linked_user,
                                            nexus_file.last_modified))

            elif key == 'RESERVOIR':
                # this is a special case for multi-reservoir files
                if fcs_file.multi_reservoir_files is None:
                    fcs_file.multi_reservoir_files = {}
                _, reservoir_name, submodel_fcs_path = (
                    fo.get_multiple_expected_sequential_values(flat_fcs_file_content[i:], 3, ['NORPT']))
                submodel_fcs_path = fo.get_full_file_path(submodel_fcs_path, origin_path)
                reservoir_name = str(reservoir_name)
                fcs_file.multi_reservoir_files[reservoir_name] = FcsNexusFile.generate_fcs_structure(
                    fcs_file_path=submodel_fcs_path, recursive=recursive)

            else:
                continue
        return fcs_file

    @staticmethod
    def fcs_keyword_map_single() -> dict[str, str]:
        """Returns mapping of fcs keywords to single file categories."""
        keyword_map = {
            'STRUCTURED_GRID': 'structured_grid_file',
            'OPTIONS': 'options_file',
            'RUNCONTROL': 'runcontrol_file',
            'OVERRIDE': 'override_file',
            'EOS_DEFAULTS': 'eos_default_file',
        }
        return keyword_map

    @staticmethod
    def fcs_keyword_map_multi() -> dict[str, str]:
        """Returns mapping of fcs keywords to file multiple file categories."""
        keyword_map = {
            'EQUIL': 'equil_files',
            'ROCK': 'rock_files',
            'RELPM': 'relperm_files',
            'SURFACE': 'surface_files',
            'WELLS': 'well_files',
            'PVT': 'pvt_files',
            'AQUIFER': 'aquifer_files',
            'HYD': 'hyd_files',
            'VALVE': 'valve_files',
            'WATER': 'water_files',
            'SEPARATOR': 'separator_files',
            'IPR': 'ipr_files',
            'GASLIFT': 'gaslift_files',
            'PUMP': 'pump_files',
            'COMPRESSOR': 'compressor_files',
            'CHOKE': 'choke_files',
            'ICD': 'icd_files',
            'ESP': 'esp_files',
            'POLYMER': 'polymer_files',
            'ADSORPTION': 'adsorption_files',
            'FLUXIN': 'flux_in_files',
            'TRACER_INIT': 'tracer_init_files',
        }
        return keyword_map

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Returns mapping of keywords to attribute definitions."""
        single_keywords = {k: (v, type(NexusFile)) for k, v in FcsNexusFile.fcs_keyword_map_single().items()}
        multi_keywords = {k: (v, type(dict)) for k, v in FcsNexusFile.fcs_keyword_map_multi().items()}
        return_dict = dict(single_keywords, **multi_keywords)
        return return_dict

    def move_model_files(self, new_file_path: str, new_include_file_location: str, overwrite_files: bool = False) -> \
            None:
        """Moves all the model files to a new location.

        Args:
            new_file_path (str): new file path for the fcs file e.g. /new_path/new_fcs_file.fcs
            new_include_file_location (str): new location for the included files either absolute or relative
            to the new fcs file path
            overwrite_files (bool): whether to overwrite the files if they already exist in the new location
        """
        # Take the original file, find which files have changed and write out those locations
        # figure out where to store the include files:
        file_directory = os.path.dirname(new_file_path)
        include_dir = new_include_file_location
        if not os.path.isabs(new_include_file_location):
            include_dir = os.path.join(file_directory, new_include_file_location)

        if not os.path.exists(include_dir):
            # make the folders if they don't already exist
            os.makedirs(include_dir)

        # Loop through all files in the model, writing out the contents
        for keyword, attr_name in self.fcs_keyword_map_single().items():
            file: None | NexusFile = getattr(self, attr_name, None)
            if file is None:
                # skip if there is no file
                continue
            include_name = os.path.join(include_dir, os.path.basename(file.location))
            file.write_to_file(include_name, write_includes=True, write_out_all_files=True,
                               overwrite_file=overwrite_files)
            self.change_file_path(include_name, keyword)

        for keyword, attr_name in self.fcs_keyword_map_multi().items():
            file_dict: None | dict[int, NexusFile] = getattr(self, attr_name, None)
            if file_dict is None or len(file_dict) == 0:
                continue
            for method_number, file in file_dict.items():
                include_name = os.path.join(include_dir, os.path.basename(file.location))
                file.write_to_file(include_name, write_includes=True, write_out_all_files=True,
                                   overwrite_file=overwrite_files)
                self.change_file_path(include_name, keyword, method_number)

        # write out the final fcs file
        self.write_to_file(new_file_path, write_includes=False, overwrite_file=overwrite_files)

    def write_out_case(self, new_file_path: str, new_include_file_location: str, case_suffix: str) -> None:
        """Writes out a new simulator with only modified files. For use with creating multiple cases from a base case.

        Args:
            new_file_path (str): new file path for the fcs file e.g. /new_path/new_fcs_file.fcs
            new_include_file_location (str): new location for the included files either absolute or relative
            to the new fcs file path
            case_suffix (str): suffix to append to the end of the file name e.g. case_1
        """

        def new_include_file_name(file_name: str) -> str:
            """Returns the new include file name based on the original file name plus the suffix provided."""
            file_name = os.path.basename(file_name)
            file_name, file_extension = os.path.splitext(file_name)
            file_name = f'{file_name}_{case_suffix}{file_extension}'
            return os.path.join(new_include_file_location, file_name)

        file_directory = os.path.dirname(new_file_path)
        include_dir = new_include_file_location
        if not os.path.isabs(new_include_file_location):
            include_dir = os.path.join(file_directory, new_include_file_location)

        if not os.path.exists(include_dir):
            # make the folders if they don't already exist
            os.makedirs(include_dir)

        # Loop through all files in the model, writing out the contents
        for keyword, attr_name in self.fcs_keyword_map_single().items():
            file: None | NexusFile = getattr(self, attr_name, None)
            if file is None:
                # skip if there is no file
                continue
            include_write_name = file.location
            if file.file_modified:
                include_write_name = new_include_file_name(file.location)
                file.write_to_file(include_write_name, write_includes=True, write_out_all_files=False)
            self.change_file_path(include_write_name, keyword)

        for keyword, attr_name in self.fcs_keyword_map_multi().items():
            file_dict: None | dict[int, NexusFile] = getattr(self, attr_name, None)
            if file_dict is None or len(file_dict) == 0:
                continue
            for method_number, file in file_dict.items():
                include_write_name = file.location
                if file.file_modified:
                    include_write_name = new_include_file_name(file.location)
                    file.write_to_file(include_write_name, write_includes=True, write_out_all_files=False)
                self.change_file_path(include_write_name, keyword, method_number)
        self.write_to_file(new_file_path, write_includes=False)

    def update_model_files(self) -> None:
        """Updates all the modified files in the model. Keeps file names and paths the same.

        Warning: this method overwrites the existing files!
        """
        # Loop through all files in the model, writing out the contents if they have been modified.
        for keyword, attr_name in self.fcs_keyword_map_single().items():
            file: None | NexusFile = getattr(self, attr_name, None)
            if file is None:
                # skip if there is no file
                continue
            if file.file_modified:
                file.write_to_file(write_includes=True, write_out_all_files=False, overwrite_file=True)

        for keyword, attr_name in self.fcs_keyword_map_multi().items():
            file_dict: None | dict[int, NexusFile] = getattr(self, attr_name, None)
            if file_dict is None or len(file_dict) == 0:
                continue
            for method_number, file in file_dict.items():
                if file.file_modified:
                    file.write_to_file(write_includes=True, write_out_all_files=False, overwrite_file=True)

    def change_file_path(self, new_file_path: str, token: str, method_number: int | None = None) -> bool:
        """Switch the file path for a new file_path based on the value of the associated keyword in the fcs.

        Args:
            new_file_path (str): file path to replace the existing file path with.
            token (str): token indicating the
            method_number (int | None): method number for the file to replace.
            If None then will only look for keywords that follow directly after a token
        Returns:
            (bool): representing whether the file was successfully changed
        """
        file_changed = False
        if self.file_content_as_list is None:
            raise ValueError('No file content to change file path on.')
        for index, line in enumerate(self.get_flat_list_str_file):
            if not nfo.check_token(token, line):
                continue

            if method_number is not None:
                token_from_file, intermediate_word, method_num_in_file, path_to_replace \
                    = fo.get_multiple_expected_sequential_values([line], 4, ['NORPT'])
                if int(method_num_in_file) != method_number:
                    continue
            else:
                path_to_replace = nfo.get_expected_token_value(token, line, self.get_flat_list_str_file)
            # edit the file as list
            file_to_edit, index_to_mod = self.find_which_include_file(flattened_index=index)
            if file_to_edit.file_content_as_list is None:
                raise ValueError(f'No content found within {file_to_edit.location}')
            file_to_edit.file_content_as_list[index_to_mod] = line.replace(path_to_replace, new_file_path)
            file_changed = True
            self._file_modified_set(file_changed)
        return file_changed

    @property
    def all_model_files(self) -> Generator[File]:
        """Generator property returning all the files linked to in the FCS file."""

        # Loop through all the individual files in the model
        individual_files_in_model = [self.restart_file, self.structured_grid_file, self.options_file,
                                     self.runcontrol_file, self.override_file, self.eos_default_file]

        for individual_file in individual_files_in_model:
            if individual_file is not None:
                yield individual_file

        # Loop through all the files linked to the model via keys e.g. WELLS set 1, WELLS set 2 etc.
        dicts_of_files_in_model = [self.well_files, self.surface_files, self.rock_files, self.relperm_files,
                                   self.pvt_files, self.water_files, self.equil_files, self.tracer_init_files,
                                   self.aquifer_files, self.hyd_files, self.valve_files, self.separator_files,
                                   self.ipr_files, self.gaslift_files, self.pump_files, self.compressor_files,
                                   self.choke_files, self.icd_files, self.esp_files, self.polymer_files,
                                   self.adsorption_files, self.flux_in_files]

        for dict_of_files in dicts_of_files_in_model:
            if dict_of_files is not None and any(dict_of_files):
                yield from dict_of_files.values()

    def get_model_files_by_filename(self, filename: str) -> list[File]:
        """Retrieves a list of files in the Nexus model matching the provided file name.

        Args:
            filename(str): file name to search for.
        """
        matching_files: list[File] = []

        for file in self.all_model_files:
            matching_files.extend(file.get_include_file_from_filename(filename=filename))

        return matching_files

    def to_string(self, dateformat: None | DateFormat = None, run_units: None | UnitSystem = None,
                  default_units: None | UnitSystem = None, description: None | str = None) -> str:
        """Writes a new FCS file contents to a string in Nexus format.

        Returns:
            str: string representation of the FCS file contents.
        """
        def print_method(method_files: None | dict[int, NexusFile], keyword: str, category: str = 'METHOD') -> str:
            """Helper function to print methods with their files."""
            prt_str = ''
            if method_files is None:
                return prt_str
            for method_number, file in method_files.items():
                prt_str += f'    {keyword} {category} {method_number} {file.location}\n'
            return prt_str

        printable_str = ''
        printable_str += 'DESC Model created with ResSimpy\n'
        if description is not None:
            printable_str += f'DESC {description}\n'
        if run_units is not None:
            printable_str += f'RUN_UNITS {run_units.value}\n'
        else:
            # if not provided print the default Nexus unit system
            printable_str += f'RUN_UNITS {UnitSystem.ENGLISH.value}\n'

        if default_units is not None:
            printable_str += f'DEFAULT_UNITS {default_units.value}\n'
        else:
            # if not provided print the default Nexus unit system
            printable_str += f'DEFAULT_UNITS {UnitSystem.ENGLISH.value}\n'

        if dateformat is not None:
            printable_str += f'DATEFORMAT {dateformat.value}\n'
        else:
            # if not provided print the default Nexus date format
            printable_str += f'DATEFORMAT {DateFormat.MM_DD_YYYY.value}\n'

        # Loop through all the individual files in the model
        printable_str += '\nGRID_FILES\n'

        if self.structured_grid_file is not None:
            printable_str += f'    STRUCTURED_GRID {self.structured_grid_file.location}\n'
        if self.options_file is not None:
            printable_str += f'    OPTIONS {self.options_file.location}\n'

        printable_str += '\nINITIALIZATION_FILES\n'
        printable_str += print_method(method_files=self.equil_files, keyword='EQUIL')
        printable_str += print_method(method_files=self.tracer_init_files, keyword='TRACER_INIT')

        printable_str += '\nROCK_FILES\n'
        rock_section_files = {'ROCK': self.rock_files, 'RELPM': self.relperm_files, 'ADSORPTION': self.adsorption_files}
        for method, method_files in rock_section_files.items():
            printable_str += print_method(method_files=method_files, keyword=method)

        printable_str += '\nPVT_FILES\n'
        pvt_section_files = {'PVT': self.pvt_files, 'WATER': self.water_files, 'POLYMER': self.polymer_files,
                             'SEPARATOR': self.separator_files}
        for method, method_files in pvt_section_files.items():
            printable_str += print_method(method_files=method_files, keyword=method)

        if self.aquifer_files is not None and len(self.aquifer_files) > 0:
            printable_str += '\nAQUIFER_FILES\n'
            printable_str += print_method(method_files=self.aquifer_files, keyword='AQUIFER')

        printable_str += '\nRECURRENT_FILES\n'
        if self.runcontrol_file is not None:
            printable_str += f'    RUNCONTROL {self.runcontrol_file.location}\n'
        printable_str += print_method(method_files=self.well_files, keyword='WELLS', category='SET')
        printable_str += print_method(method_files=self.surface_files, keyword='SURFACE', category='NETWORK')
        printable_str += print_method(method_files=self.ipr_files, keyword='IPR')
        if self.override_file is not None:
            printable_str += f'OVERRIDE {self.override_file.location}\n'

        net_methods = {'VALVE': self.valve_files,
                       'GASLIFT': self.gaslift_files, 'PUMP': self.pump_files,
                       'COMPRESSOR': self.compressor_files, 'CHOKE': self.choke_files,
                       'ICD': self.icd_files, 'ESP': self.esp_files, 'HYD': self.hyd_files}
        if any(x for x in net_methods.values()):
            printable_str += '\nNET_METHOD_FILES\n'
            for method, method_files in net_methods.items():
                printable_str += print_method(method_files=method_files, keyword=method)

        return printable_str

    def _add_file(self, file: NexusFile, keyword: str, method_number: int | None = None) -> None:
        """Adds a file to the FCS file under the specified keyword and method number.

        Args:
            file (NexusFile): The NexusFile to add.
            keyword (str): The keyword under which to add the file.
            method_number (int | None): The method number for the file. If None, adds as a single file.
        """
        if method_number is not None:
            if not hasattr(self, self.fcs_keyword_map_multi()[keyword]):
                setattr(self, self.fcs_keyword_map_multi()[keyword], {})
            getattr(self, self.fcs_keyword_map_multi()[keyword])[method_number] = file
        else:
            setattr(self, self.fcs_keyword_map_single()[keyword], file)

        if self.include_objects is not None:
            if isinstance(self.include_objects, list):
                self.include_objects.append(file)
        if self.include_locations is not None and file.location not in self.include_locations:
            self.include_locations.append(file.location)
        self.files_info.append((file.location, file.linked_user, file.last_modified))
